# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=broad-except
"""Path check utility functions."""

from pathlib import Path
from typing import Optional, Union


def _check_local_path(string: str) -> Optional[Path]:
    """Check if a string is a local path.

    Parameters
    ----------
    string : str
        The string to check.

    Returns
    -------
    bool
        True if the path is a local path.
    """
    try:
        path = Path(string).resolve()
    except BaseException:  # pragma: no cover
        return None
    if path.exists():
        return path
    return None


def _get_raw_path_string(path: Union[str, Path]) -> str:
    """Get the raw path string.

    Parameters
    ----------
    path : Union[str, Path]
        The string to check.

    Returns
    -------
    str
        The raw path string.
    """
    if not isinstance(path, str):
        path = str(path)
    while path.startswith('r"') and path.endswith('"'):
        path = path[2:-1]
    return f'r"{path}"'


def get_path_string(path: str) -> str:
    """Get the path string.

    Parameters
    ----------
    path : str
        The string to check.

    Returns
    -------
    str
        The local path string.
    """
    resolved = _check_local_path(path)
    if not resolved:
        return _get_raw_path_string(path)
    return _get_raw_path_string(resolved)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez agents model."""

from typing import Iterator, List

from pydantic import Field, model_validator
from typing_extensions import Annotated, Self

from ..common import WaldiezBase
from .agent import WaldiezAgent
from .assistant import WaldiezAssistant
from .group_manager.group_manager import WaldiezGroupManager
from .rag_user import WaldiezRagUser
from .reasoning import WaldiezReasoningAgent
from .swarm_agent import WaldiezSwarmAgent
from .user_proxy import WaldiezUserProxy


class WaldiezAgents(WaldiezBase):
    """Waldiez agents model.

    Attributes
    ----------
    users : List[WaldiezUserProxy]
        User proxy agents.
    assistants : List[WaldiezAssistant]
        Assistant agents.
    managers : List[WaldiezGroupManager]
        Group chat managers.
    rag_users : List[WaldiezRagUser]
        RAG user agents.
    """

    users: Annotated[
        List[WaldiezUserProxy],
        Field(
            title="Users.",
            description="User proxy agents",
            default_factory=list,
        ),
    ]
    assistants: Annotated[
        List[WaldiezAssistant],
        Field(
            title="Assistants.",
            description="Assistant agents",
            default_factory=list,
        ),
    ]
    managers: Annotated[
        List[WaldiezGroupManager],
        Field(
            title="Managers.",
            description="Group chat managers",
            default_factory=list,
        ),
    ]
    rag_users: Annotated[
        List[WaldiezRagUser],
        Field(
            title="RAG Users.",
            description="RAG user agents",
            default_factory=list,
        ),
    ]
    swarm_agents: Annotated[
        List[WaldiezSwarmAgent],
        Field(
            title="Swarm Agents.",
            description="Swarm agents",
            default_factory=list,
        ),
    ]
    reasoning_agents: Annotated[
        List[WaldiezReasoningAgent],
        Field(
            title="Reasoning Agents.",
            description="Reasoning agents",
            default_factory=list,
        ),
    ]

    @property
    def members(self) -> Iterator[WaldiezAgent]:
        """Get all agents.

        Yields
        ------
        WaldiezAgent
            The agents.
        """
        yield from self.users
        yield from self.assistants
        yield from self.rag_users
        yield from self.reasoning_agents
        yield from self.swarm_agents
        yield from self.managers

    @model_validator(mode="after")
    def validate_agents(self) -> Self:
        """Validate the agents.

        - At least two agents are required.
        - All the agent IDs must be unique.

        Returns
        -------
        WaldiezAgents
            The agents.

        Raises
        ------
        ValueError
            If the agents are invalid.
        """
        all_agent_ids = [agent.id for agent in self.members]
        if len(all_agent_ids) < 2:
            raise ValueError("At least two agents are required.")
        if len(all_agent_ids) != len(set(all_agent_ids)):
            raise ValueError("Agent IDs must be unique.")
        return self

    def validate_flow(self, model_ids: List[str], skill_ids: List[str]) -> None:
        """Validate the flow of the agents.

        - Validate the linked models (the referenced model ids must exist).
        - Validate the linked skills (the referenced skill ids must exist).
        - Validate the code execution (the referenced functions must exist).

        Parameters
        ----------
        model_ids : List[str]
            The list of model IDs.
        skill_ids : List[str]
            The list of skill IDs.

        Raises
        ------
        ValueError
            If the flow is invalid.
        """
        all_agent_ids = [agent.id for agent in self.members]
        for agent in self.members:
            agent.validate_linked_models(model_ids)
            agent.validate_linked_skills(skill_ids, agent_ids=all_agent_ids)
            agent.validate_code_execution(skill_ids=skill_ids)
            if agent.agent_type == "manager" and isinstance(
                agent, WaldiezGroupManager
            ):
                agent.validate_transitions(agent_ids=all_agent_ids)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Base class to inherit from."""

from typing import Any, Dict

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class WaldiezBase(BaseModel):
    """Base model class to inherit from.

    It contains the default configuration for all models.
    It also `model_dumps` by alias by default.
    """

    model_config = ConfigDict(
        extra="ignore",
        # treat `skillId` as `skill_id`
        alias_generator=to_camel,
        # allow passing either `skill_id` or `skillId`
        populate_by_name=True,
        # allow setting any attribute after initialization
        frozen=False,
    )

    def model_dump(self, **kwargs: Any) -> Dict[str, Any]:
        """Dump the model to a dictionary.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Dict[str, Any]
            The dictionary representation of the model.
        """
        by_alias = kwargs.pop("by_alias", None)
        if by_alias is None:
            by_alias = True
        if not isinstance(by_alias, bool):
            by_alias = True
        return super().model_dump(by_alias=by_alias, **kwargs)

    def model_dump_json(self, **kwargs: Any) -> str:
        """Dump the model to a JSON string.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        str
            The JSON string.
        """
        by_alias = kwargs.pop("by_alias", None)
        if by_alias is None:
            by_alias = True
        if not isinstance(by_alias, bool):
            by_alias = True
        return super().model_dump_json(by_alias=by_alias, **kwargs)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Nested chat model."""

from typing import Any, Optional

from pydantic import Field, field_validator, model_validator
from typing_extensions import Annotated, Self

from ..common import WaldiezBase
from .chat_message import WaldiezChatMessage

NESTED_CHAT_MESSAGE = "nested_chat_message"
NESTED_CHAT_REPLY = "nested_chat_reply"
NESTED_CHAT_ARGS = ["recipient", "messages", "sender", "config"]
NESTED_CHAT_TYPES = (
    [
        "ConversableAgent",
        "List[Dict[str, Any]]",
        "ConversableAgent",
        "Dict[str, Any]",
    ],
    "Union[Dict[str, Any], str]",
)


class WaldiezChatNested(WaldiezBase):
    """Nested chat class.

    Attributes
    ----------
    message : WaldiezChatMessage
        The message in a nested chat (sender -> recipient).
    reply : WaldiezChatMessage
        The reply in a nested chat (recipient -> sender).
    """

    message: Annotated[
        Optional[WaldiezChatMessage],
        Field(
            None,
            title="Message",
            description="The message in a nested chat (sender -> recipient).",
        ),
    ]
    reply: Annotated[
        Optional[WaldiezChatMessage],
        Field(
            None,
            title="Reply",
            description="The reply in a nested chat (recipient -> sender).",
        ),
    ]

    _message_content: Optional[str] = None
    _reply_content: Optional[str] = None

    @property
    def message_content(self) -> Optional[str]:
        """Get the message content."""
        return self._message_content

    @property
    def reply_content(self) -> Optional[str]:
        """Get the reply content."""
        return self._reply_content

    @field_validator("message", "reply", mode="before")
    @classmethod
    def validate_message(cls, value: Any) -> WaldiezChatMessage:
        """Validate the message.

        Parameters
        ----------
        value : Any
            The value.

        Returns
        -------
        WaldiezChatMessage
            The validated message.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if not value:
            return WaldiezChatMessage(
                type="none", use_carryover=False, content=None, context={}
            )
        if isinstance(value, str):
            return WaldiezChatMessage(
                type="string", use_carryover=False, content=value, context={}
            )
        if isinstance(value, dict):
            return WaldiezChatMessage.model_validate(value)
        if isinstance(value, WaldiezChatMessage):
            return value
        raise ValueError(f"Invalid message type: {type(value)}")

    @model_validator(mode="after")
    def validate_nested_chat(self) -> Self:
        """Validate the nested chat.

        Returns
        -------
        WaldiezChatNested
            The validated nested chat.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if self.message is not None:
            self._message_content = self.message.content_body
            if self.message.type == "none":
                self._message_content = ""
            if self.message.type == "string":
                self._message_content = self.message.content
            if self.message.type == "method":
                self._message_content = self.message.validate_method(
                    function_name=NESTED_CHAT_MESSAGE,
                    function_args=NESTED_CHAT_ARGS,
                )

        if self.reply is not None:
            self._reply_content = self.reply.content_body
            if self.reply.type == "none":
                self._reply_content = ""
            if self.reply.type == "string":
                self._reply_content = self.reply.content
            if self.reply.type == "method":
                self._reply_content = self.reply.validate_method(
                    function_name=NESTED_CHAT_REPLY,
                    function_args=NESTED_CHAT_ARGS,
                )
        return self



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Common utilities for exporting chats."""

from typing import Any, Callable, Dict, Optional, Tuple

from waldiez.models import WaldiezAgent, WaldiezChat


def update_summary_chat_args(
    chat_args: Dict[str, Any],
    string_escape: Callable[[str], str],
) -> Dict[str, Any]:
    """Escape quotes in the summary args if they are strings.

    Parameters
    ----------
    chat_args : Dict[str, Any]
        The chat arguments.
    string_escape : Callable[[str], str]
        The function to escape the string

    Returns
    -------
    Dict[str, Any]
        The chat arguments with the summary prompt escaped.
    """
    if "summary_args" in chat_args and isinstance(
        chat_args["summary_args"], dict
    ):
        for key, value in chat_args["summary_args"].items():
            if isinstance(value, str):
                chat_args["summary_args"][key] = string_escape(value)
    return chat_args


def get_chat_message_string(
    sender: WaldiezAgent,
    chat: WaldiezChat,
    chat_names: Dict[str, str],
    string_escape: Callable[[str], str],
) -> Tuple[str, Optional[str]]:
    """Get the agent's message as a string.

    Parameters
    ----------
    sender : WaldiezAgent
        The sender.
    chat : WaldiezChat
        The chat.
    chat_names : Dict[str, str]
        A mapping of chat id to chat name with all the chats in the flow.
    string_escape : Callable[[str], str]
        The function to escape the string.

    Returns
    -------
    Tuple[str, Optional[str]]
        If the message is a string, the message content and None.
        If the message is a method, the method name and the method content.
        If the message is None, 'None' and None.
    """
    if (
        not chat.message
        or chat.message.type == "none"
        or chat.message.content is None
        or chat.message_content is None
    ):
        return "None", None
    if chat.message.type == "string":
        return string_escape(chat.message.content), None

    is_rag_with_carryover = (
        sender.agent_type == "rag_user" and chat.message.use_carryover
    )
    chat_name = chat_names[chat.id]
    function_content, function_name = chat.get_message_function(
        name_suffix=chat_name,
        is_rag=is_rag_with_carryover,
    )
    return function_name, function_content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Handle agent contents before and after the agent(s) exports."""

from typing import List, Tuple, Union

from waldiez.exporting.base import (
    AgentPosition,
    AgentPositions,
    ExporterReturnType,
    ExportPosition,
)
from waldiez.models import WaldiezAgent


def add_after_all_agents_content(
    agents_contents: str,
    after_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
) -> str:
    """Add the after all agents content.

    Parameters
    ----------
    agents_contents : str
        The agents content.
    after_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        The after export.

    Returns
    -------
    str
        The agents content with the after all agents content.
    """
    new_content = str(agents_contents)
    # let's get the list first, and sort it by the order
    after_all_agent_exports = [
        (content, position)
        for content, position in after_export
        if isinstance(position, AgentPosition)
        and position.position == AgentPositions.AFTER_ALL
    ]
    by_order = sorted(after_all_agent_exports, key=lambda x: x[1].order)
    for content, _ in by_order:
        new_content += content + "\n"
    if not new_content.endswith("\n"):
        new_content += "\n"
    return new_content


def add_before_all_agents_content(
    agents_contents: str,
    before_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
) -> str:
    """Add the before all agents content.

    Parameters
    ----------
    agents_contents : str
        The agents content.
    before_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        The before export.

    Returns
    -------
    str
        The agents content with the before all agents content.
    """
    new_content = str(agents_contents)
    before_all_agents_exports = [
        (content, position)
        for content, position in before_export
        if isinstance(position, AgentPosition)
        and position.position == AgentPositions.BEFORE_ALL
    ]
    for content, _ in before_all_agents_exports:
        new_content = content + "\n" + new_content
    if not new_content.startswith("\n"):
        new_content = "\n" + new_content
    return new_content


def add_before_agent_content(
    agent_content: str,
    before_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
    agent: WaldiezAgent,
) -> str:
    """Add the before agent content.

    Parameters
    ----------
    agent_content : str
        The agent content.
    before_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        The before export.
    agent : WaldiezAgent
        The agent.

    Returns
    -------
    str
        The agent content with the before agent content.
    """
    new_content = str(agent_content)
    for content, position in before_export:
        if (
            isinstance(position, AgentPosition)
            and position.agent == agent
            and position.position == AgentPositions.BEFORE
        ):
            new_content = content + "\n" + new_content
    if not new_content.startswith("\n"):
        new_content = "\n" + new_content
    return new_content


def add_after_agent_content(
    agent_content: str,
    after_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
    agent: WaldiezAgent,
) -> str:
    """Add the after agent content.

    Parameters
    ----------
    agent_content : str
        The agent content.
    after_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        The after export.
    agent : WaldiezAgent
        The agent.

    Returns
    -------
    str
        The agent content with the after agent content.
    """
    new_content = str(agent_content)
    for content, position in after_export:
        if (
            isinstance(position, AgentPosition)
            and position.agent == agent
            and position.position == AgentPositions.AFTER
        ):
            new_content += content + "\n"
    if not new_content.endswith("\n"):
        new_content += "\n"
    return new_content


def gather_agent_outputs(
    before_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
    after_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
    agent_outputs: List[ExporterReturnType],
) -> ExporterReturnType:
    """Gather all the agent outputs.

    Parameters
    ----------
    before_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        The before export.
    after_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        The after export.
    agent_outputs : List[ExporterReturnType]
        The agent outputs.

    Returns
    -------
    ExporterReturnType
        The gathered agent outputs.
    """
    agents_contents = ""
    agents_imports = []
    agents_before_export = []
    agents_after_export = []
    agents_env_vars = []
    for output in agent_outputs:
        if output["content"]:
            agents_contents += output["content"]
        if output["imports"]:
            agents_imports.extend(output["imports"])
        if output["before_export"]:
            agents_before_export.extend(output["before_export"])
        if output["after_export"]:
            agents_after_export.extend(output["after_export"])
        if output["environment_variables"]:
            agents_env_vars.extend(output["environment_variables"])
    agents_contents = add_before_all_agents_content(
        agents_contents,
        before_export,
    )
    agents_contents = add_after_all_agents_content(
        agents_contents,
        after_export,
    )
    agents_contents = agents_contents.replace("\n\n\n\n", "\n\n\n")
    while agents_contents.endswith("\n\n\n"):
        agents_contents = agents_contents[: -len("\n")]
    return {
        "content": agents_contents,
        "imports": agents_imports,
        "before_export": agents_before_export,
        "after_export": agents_after_export,
        "environment_variables": agents_env_vars,
    }



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa: E501
# pylint: disable=inconsistent-quotes, line-too-long
"""Get the main function."""


def get_def_main(flow_chats: str, after_run: str, is_async: bool) -> str:
    """Get the main function.

    When exporting to python, waldiez_chats string will be the
    content of the main function. It contains either a
    `{sender.initiate_chat(recipient, ...)}` (if there is only one chat)
    or `initiate_chats([..])`, with the list of chats to initiate.
    If async: (sender.a_initiate_chat, a_initiate_chats)

    Parameters
    ----------
    flow_chats : str
        The content of the main function.
    after_run : str
        The content after the run of the flow.
    is_async : bool
        Whether the main function is asynchronous.
    Returns
    -------
    str
        The main function.
    """
    content = ""
    if is_async:
        content += "async "
    content += "def main():\n"
    content += "    # type: () -> Union[ChatResult, List[ChatResult], Dict[int, ChatResult]]\n"
    content += '    """Start chatting."""\n'
    content += f"{flow_chats}" + "\n"
    if is_async:
        content += "    await stop_logging()"
    else:
        content += "    stop_logging()"
    content += after_run
    content += "\n    return results\n\n\n"
    if is_async:
        content += "async def call_main():\n"
    else:
        content += "def call_main() -> None:\n"
    content += '    """Run the main function and print the results."""\n'
    content += "    results: Union[ChatResult, List[ChatResult], Dict[int, ChatResult]] = "
    if is_async:
        content += "await "
    content += "main()\n"
    content += "    if isinstance(results, dict):\n"
    content += "        # order by key\n"
    content += "        ordered_results = dict(sorted(results.items()))\n"
    content += "        for _, result in ordered_results.items():\n"
    content += "            pprint(asdict(result))\n"
    content += "    elif isinstance(results, list):\n"
    content += "        for result in results:\n"
    content += "            pprint(asdict(result))\n"
    content += "    else:\n"
    content += "        pprint(asdict(results))\n"
    content += 'if __name__ == "__main__":\n'
    if is_async:
        content += "    anyio.run(call_main)\n"
    else:
        content += "    call_main()\n"
    return content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Reasoning agent model."""

from .reasoning_agent import WaldiezReasoningAgent
from .reasoning_agent_data import WaldiezReasoningAgentData
from .reasoning_agent_reason_config import WaldiezReasoningAgentReasonConfig

__all__ = [
    "WaldiezReasoningAgentReasonConfig",
    "WaldiezReasoningAgent",
    "WaldiezReasoningAgentData",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Skill related models."""

from .skill import SHARED_SKILL_NAME, WaldiezSkill
from .skill_data import WaldiezSkillData

__all__ = [
    "SHARED_SKILL_NAME",
    "WaldiezSkill",
    "WaldiezSkillData",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=import-outside-toplevel,reimported
"""Environment related utilities."""

import importlib.util
import os
import site
import sys
import warnings
from typing import Dict, List, Tuple


def in_virtualenv() -> bool:
    """Check if we are inside a virtualenv.

    Returns
    -------
    bool
        True if inside a virtualenv, False otherwise.
    """
    return hasattr(sys, "real_prefix") or (
        hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
    )


def refresh_environment() -> None:
    """Refresh the environment."""
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            module="flaml",
            message="^.*flaml.automl is not available.*$",
        )
        from autogen.io import IOStream  # type: ignore

        default_io_stream = IOStream.get_default()
        site.main()
        # pylint: disable=import-outside-toplevel
        modules_to_reload = [mod for mod in sys.modules if "autogen" in mod]
        for mod in modules_to_reload:
            del sys.modules[mod]
        import autogen  # type: ignore
        from autogen.io import IOStream

        importlib.reload(autogen)
        # restore the default IOStream
        IOStream.set_global_default(default_io_stream)
        # reload any other modules that may have been affected
        for mod in modules_to_reload:
            if mod not in sys.modules:
                importlib.import_module(mod)


def set_env_vars(flow_env_vars: List[Tuple[str, str]]) -> Dict[str, str]:
    """Set environment variables and return the old ones (if any).

    Parameters
    ----------
    flow_env_vars : List[Tuple[str, str]]
        The environment variables to set.

    Returns
    -------
    Dict[str, str]
        The old environment variables.
    """
    old_vars: Dict[str, str] = {}
    for var_key, var_value in flow_env_vars:
        if var_key:
            current = os.environ.get(var_key, "")
            old_vars[var_key] = current
            os.environ[var_key] = var_value
    return old_vars


def reset_env_vars(old_vars: Dict[str, str]) -> None:
    """Reset the environment variables.

    Parameters
    ----------
    old_vars : Dict[str, str]
        The old environment variables.
    """
    for var_key, var_value in old_vars.items():
        if not var_value:
            os.environ.pop(var_key, "")
        else:
            os.environ[var_key] = var_value



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""RAG User Agent related string generation."""

from .rag_user import get_rag_user_extras, get_rag_user_retrieve_config_str

__all__ = ["get_rag_user_retrieve_config_str", "get_rag_user_extras"]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=unused-argument
"""Get the extras for a swarm agent."""

from typing import Callable, Dict, List, Tuple

from waldiez.exporting.chats.utils.nested import get_nested_chat_queue
from waldiez.models import (
    WaldiezAgent,
    WaldiezChat,
    WaldiezSwarmAfterWork,
    WaldiezSwarmAgent,
    WaldiezSwarmOnCondition,
    WaldiezSwarmUpdateSystemMessage,
)

# SwarmAgent is a subclass of ConversableAgent.

# Additional args:
# functions (List[Callable]):
#   -A list of functions to register with the agent.
# update_agent_state_before_reply (List[Callable]):
# - A list of functions, including UPDATE_SYSTEM_MESSAGEs,
#   called to update the agent before it replies.

# Additional methods:
# register_hand_off(hand_offs: List[AfterWork|OnCondition]):


def get_swarm_extras(
    agent: WaldiezAgent,
    agent_names: Dict[str, str],
    skill_names: Dict[str, str],
    chats: Tuple[List[WaldiezChat], Dict[str, str]],
    is_async: bool,
    serializer: Callable[..., str],
    string_escape: Callable[[str], str],
) -> Tuple[str, str, str]:
    """Get the extras of a swarm agent.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent to get the extras for.
    agent_names : Dict[str, str]
        A mapping of agent IDs to agent names.
    skill_names : Dict[str, str]
        A mapping of skill IDs to skill names.
    chats : Tuple[List[WaldiezChat], Dict[str, str]]
        The list of all chats and the mapping of chat IDs to chat names.
    is_async : bool
        Whether the chat is asynchronous.
    serializer : Callable[..., str]
        The serializer to get the string representation of an object.
    string_escape : Callable[[str], str]
        The function to escape the string quotes and newlines.
    Returns
    -------
    Tuple[str, str, str]
        The extras of the swarm agent:
        the content before the agent,
        the extra argument(s) for the agent initialization,
        and the content after the agent.
    """
    args_string = ""
    before_agent = ""
    after_agent = ""
    if agent.agent_type != "swarm" or not isinstance(agent, WaldiezSwarmAgent):
        return args_string, before_agent, after_agent
    args_string = get_function_arg(agent, skill_names)
    before_reply = get_update_agent_state_before_reply_arg(
        agent=agent,
        agent_names=agent_names,
        skill_names=skill_names,
        string_escape=string_escape,
    )
    args_string += before_reply[0]
    before_agent += before_reply[1]
    before_registration, after_agent = get_agent_handoff_registrations(
        agent=agent,
        agent_names=agent_names,
        all_chats=chats[0],
        chat_names=chats[1],
        is_async=is_async,
        serializer=serializer,
        string_escape=string_escape,
    )
    before_agent += before_registration
    return before_agent, args_string, after_agent


def get_function_arg(
    agent: WaldiezSwarmAgent,
    skill_names: Dict[str, str],
) -> str:
    """Get the function argument of a swarm agent.

    Parameters
    ----------
    agent : WaldiezSwarmAgent
        The swarm agent to get the function argument for.
    skill_names : Dict[str, str]
        A mapping of skill IDs to skill names.

    Returns
    -------
    str
        The function argument of the swarm agent.
    """
    tab = "    "
    arg_string = f"{tab}functions=["
    added_skills = False
    for function in agent.data.functions:
        skill_name = skill_names.get(function, "")
        if skill_name:
            arg_string += "\n" + f"{tab}{tab}{skill_name},"
            added_skills = True
    if added_skills:
        arg_string += "\n" + tab
    arg_string += "],\n"
    return arg_string


def get_update_agent_state_before_reply_arg(
    agent: WaldiezSwarmAgent,
    agent_names: Dict[str, str],
    skill_names: Dict[str, str],
    string_escape: Callable[[str], str],
) -> Tuple[str, str]:
    """Get the update_agent_state_before_reply argument of a swarm agent.

    Parameters
    ----------
    agent : WaldiezSwarmAgent
        The swarm agent to get the argument for.
    agent_names : Dict[str, str]
        A mapping of agent IDs to agent names.
    skill_names : Dict[str, str]
        A mapping of skill IDs to skill names.
    string_escape : Callable[[str], str]
        The function to escape the string quotes and newlines.

    Returns
    -------
    Tuple[str, str]
        The update_agent_state_before_reply argument of the swarm agent
        and the content before the agent if any.
    """
    #     update_function_type : Literal["string", "callable"]
    #     The type of the update function. Can be either a string or a callable.
    # update_function : str
    #     "The string template or function definition to update "
    #          "the agent's system message. Can be a string or a Callable. "
    #           "If the function_type is 'string' it will be used as a "
    #            "template and substitute the context variables. "
    #         ag2 checks for: vars = re.findall(r"\{(\w+)\}", function)
    #     "If function_type is 'callable', it should have signature:
    #  "def custom_update_system_message("
    #     " agent: ConversableAgent, "
    #     " messages: List[Dict[str, Any]]
    #   ) -> str"
    tab = "    "
    before_agent = ""
    arg_string = f"{tab}update_agent_state_before_reply=["
    added_functions = False
    # pylint: disable=line-too-long
    for function in agent.data.update_agent_state_before_reply:
        if isinstance(function, WaldiezSwarmUpdateSystemMessage):
            added_functions = True
            if function.update_function_type == "callable":
                function_content, function_name = function.get_update_function(
                    name_suffix=agent_names[agent.id],
                )
                arg_string += (
                    "\n" + f"{tab}{tab}UPDATE_SYSTEM_MESSAGE({function_name}),"
                )
                before_agent += "\n" + function_content + "\n"
            else:
                escaped_function = string_escape(function.update_function)
                arg_string += (
                    "\n"
                    + f'{tab}{tab}UPDATE_SYSTEM_MESSAGE("{escaped_function}"),'
                )
        else:
            skill_name = skill_names.get(function, "")
            if skill_name:
                added_functions = True
                arg_string += "\n" + f"{tab}{tab}{skill_name},"
    if added_functions:
        arg_string = arg_string + "\n" + tab
    arg_string += "],\n"
    return arg_string, before_agent


def get_agent_handoff_registrations(
    agent: WaldiezSwarmAgent,
    agent_names: Dict[str, str],
    all_chats: List[WaldiezChat],
    chat_names: Dict[str, str],
    is_async: bool,
    serializer: Callable[..., str],
    string_escape: Callable[[str], str],
) -> Tuple[str, str]:
    """Get the agent handoff registrations of a swarm agent.

    Parameters
    ----------
    agent : WaldiezSwarmAgent
        The swarm agent to get the agent handoff registrations for.
    agent_names : Dict[str, str]
        A mapping of agent IDs to agent names.
    all_chats : List[WaldiezChat]
        The list of all chats.
    chat_names : Dict[str, str]
        A mapping of chat IDs to chat names.
    is_async : bool
        Whether the chat is asynchronous.
    serializer : Callable[..., str]
        The serializer to get the string representation of an object.
    string_escape : Callable[[str], str]
        The function to escape the string quotes and newlines.

    Returns
    -------
    Tuple[str, str]
        the contents before and after the agent.
    """
    agent_name = agent_names[agent.id]
    registrations = []
    before_agent = ""
    after_agent = ""
    if not agent.handoffs:
        return before_agent, after_agent
    tab = "    "
    after_agent = f"{agent_name}.register_hand_off(" + "\n" + f"{tab}[" + "\n"
    for hand_off in agent.handoffs:
        if isinstance(hand_off, WaldiezSwarmOnCondition):
            registration, before_handoff = get_agent_on_condition_handoff(
                agent=agent,
                hand_off=hand_off,
                agent_names=agent_names,
                all_chats=all_chats,
                chat_names=chat_names,
                is_async=is_async,
                serializer=serializer,
                string_escape=string_escape,
            )
            if registration:
                registrations.append(registration)
            before_agent += before_handoff
        elif isinstance(hand_off, WaldiezSwarmAfterWork):
            registration, before_handoff = get_agent_after_work_handoff(
                hand_off=hand_off,
                agent_names=agent_names,
                agent_name=agent_name,
            )
            registrations.append(registration)
            before_agent += before_handoff
    after_agent += "\n".join(registrations) + "\n" + f"{tab}]" + "\n" + ")"
    return before_agent, after_agent


def get_agent_after_work_handoff(
    hand_off: WaldiezSwarmAfterWork,
    agent_names: Dict[str, str],
    agent_name: str,
) -> Tuple[str, str]:
    """Get the agent's after work hand off registration.

    Parameters
    ----------
    hand_off : WaldiezSwarmAfterWork
        The hand off to get the registration for.
    agent_names : Dict[str, str]
        A mapping of agent IDs to agent names.
    agent_name : str
        The name of the agent to register the hand off.

    Returns
    -------
    Tuple[str, str]
        The registration and the content before the agent.
    """
    before_agent = ""
    tab = "    "
    recipient_type = hand_off.recipient_type
    recipient, function_content = hand_off.get_recipient(
        agent_names=agent_names,
        name_suffix=agent_name,
    )
    registration = f"{tab}{tab}{recipient},"
    if recipient_type == "callable" and function_content:
        before_agent += "\n" + function_content + "\n"
    return registration, before_agent


def get_agent_on_condition_handoff(
    agent: WaldiezSwarmAgent,
    hand_off: WaldiezSwarmOnCondition,
    agent_names: Dict[str, str],
    all_chats: List[WaldiezChat],
    chat_names: Dict[str, str],
    is_async: bool,
    serializer: Callable[..., str],
    string_escape: Callable[[str], str],
) -> Tuple[str, str]:
    """Get the agent's on condition hand off registration.

    Parameters
    ----------
    agent : WaldiezSwarmAgent
        The agent to get the registration for.
    hand_off : WaldiezSwarmAfterWork
        The hand off to get the registration for.
    agent_names : Dict[str, str]
        A mapping of agent IDs to agent names.
    all_chats : List[WaldiezChat]
        The list of all chats.
    chat_names : Dict[str, str]
        A mapping of chat IDs to chat names.
    is_async : bool
        Whether the chat is asynchronous.
    serializer : Callable[..., str]
        The serializer to get the string representation of an object.
    string_escape : Callable[[str], str]
        The function to escape the string quotes and newlines.

    Returns
    -------
    Tuple[str, str]
        The registration and the content before the agent.
    """
    before_agent = ""
    registration = ""
    available, available_function = hand_off.get_available(
        name_suffix=agent_names[agent.id],
    )
    if available and not available_function:
        available = f'"{string_escape(available)}"'
    if hand_off.target_type == "agent":
        recipient = agent_names[hand_off.target.id]
        condition = (
            string_escape(hand_off.condition) or f"Transfer to {recipient}"
        )
        results = _get_agent_on_condition_handoff_to_agent(
            recipient=recipient,
            available=available,
            condition=condition,
            available_function=available_function,
        )
        before_agent += results[0]
        registration = results[1]
    # else: # target_type == "nested_chat"
    if hand_off.target_type == "nested_chat":
        condition = _get_condition_string(
            condition=hand_off.condition,
            chat_id=hand_off.target.id,
            all_chats=all_chats,
            agent_names=agent_names,
        )
        results = _get_agent_on_condition_handoff_to_nested_chat(
            agent=agent,
            agent_names=agent_names,
            condition=hand_off.condition,
            available=available,
            available_function=available_function,
            all_chats=all_chats,
            chat_names=chat_names,
            is_async=is_async,
            serializer=serializer,
            string_escape=string_escape,
        )
        before_agent += results[0]
        registration = results[1]
    return registration, before_agent


def _get_agent_on_condition_handoff_to_agent(
    recipient: str,
    available: str,
    condition: str,
    available_function: str,
) -> Tuple[str, str]:
    before_agent = ""
    tab = "    "
    on_condition = (
        f"{tab}{tab}ON_CONDITION(" + "\n"
        f"{tab}{tab}{tab}target={recipient}," + "\n"
        f'{tab}{tab}{tab}condition="{condition}",' + "\n"
    )
    if available:
        on_condition += f"{tab}{tab}{tab}available={available}," + "\n"
    if available_function:
        before_agent += "\n" + available_function + "\n"
    on_condition += f"{tab}{tab}),"
    return before_agent, on_condition


# pylint: disable=too-many-locals
def _get_agent_on_condition_handoff_to_nested_chat(
    agent: WaldiezAgent,
    agent_names: Dict[str, str],
    condition: str,
    available: str,
    available_function: str,
    all_chats: List[WaldiezChat],
    chat_names: Dict[str, str],
    is_async: bool,
    serializer: Callable[..., str],
    string_escape: Callable[[str], str],
) -> Tuple[str, str]:
    if not agent.data.nested_chats or not agent.data.nested_chats[0].messages:
        return "", ""
    chat_queue, extra_methods = get_nested_chat_queue(
        nested_chat=agent.data.nested_chats[0],
        agent=agent,
        agent_names=agent_names,
        chat_names=chat_names,
        all_chats=all_chats,
        serializer=serializer,
        string_escape=string_escape,
    )
    if not chat_queue:
        return "", ""
    before_agent = ""
    tab = "    "
    chat_queue_var_name = f"{agent_names[agent.id]}_handoff_nested_chat_queue"
    if extra_methods:
        before_agent += "\n".join(extra_methods) + "\n"
    before_agent += f"{chat_queue_var_name} = {chat_queue} " + "\n"
    condition_string = string_escape(condition)
    on_condition = (
        f"{tab}{tab}ON_CONDITION(" + "\n"
        f"{tab}{tab}{tab}target=" + "{\n"
        f'{tab}{tab}{tab}{tab}"chat_queue": {chat_queue_var_name},' + "\n"
        f'{tab}{tab}{tab}{tab}"config": None,' + "\n"
        f'{tab}{tab}{tab}{tab}"reply_func_from_nested_chats": None,' + "\n"
        f'{tab}{tab}{tab}{tab}"use_async": {is_async},' + "\n"
        f"{tab}{tab}{tab}" + "},\n"
        f'{tab}{tab}{tab}condition="{condition_string}",' + "\n"
    )
    if available:
        on_condition += f"{tab}{tab}{tab}available={available}," + "\n"
    if available_function:
        before_agent += "\n" + available_function + "\n"
    on_condition += f"{tab}{tab}),"
    return before_agent, on_condition


def _get_condition_string(
    condition: str,
    chat_id: str,
    all_chats: List[WaldiezChat],
    agent_names: Dict[str, str],
) -> str:
    if not condition:
        chat = next((c for c in all_chats if c.id == chat_id), None)
        if chat:
            target_name = agent_names[chat.target]
            return f"Transfer to {target_name}"
        return "Transfer to the next agent"
    return condition



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez RAG user agent data."""

from pydantic import Field
from typing_extensions import Annotated

from ..user_proxy import WaldiezUserProxyData
from .retrieve_config import WaldiezRagUserRetrieveConfig


class WaldiezRagUserData(WaldiezUserProxyData):
    """RAG user agent data.

    The data for a RAG user agent.

    Attributes
    ----------
    use_message_generator: bool
        Whether to use the message generator in user's chats. Defaults to False.
    retrieve_config : WaldiezRagUserRetrieveConfig
        The RAG user agent's retrieve config.

    """

    retrieve_config: Annotated[
        WaldiezRagUserRetrieveConfig,
        Field(
            title="Retrieve Config",
            description="The RAG user agent's retrieve config",
            default_factory=WaldiezRagUserRetrieveConfig,
            alias="retrieveConfig",
        ),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Skill model."""

from typing import Dict, List

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from ..common import WaldiezBase, get_function, now, parse_code_string
from .skill_data import WaldiezSkillData

SHARED_SKILL_NAME = "waldiez_shared"


class WaldiezSkill(WaldiezBase):
    """Waldiez Skill.

    Attributes
    ----------
    id : str
        The ID of the skill.
    type : Literal["skill"]
        The type of the "node" in a graph: "skill".
    name : str
        The name of the skill.
    description : str
        The description of the skill.
    tags : List[str]
        The tags of the skill.
    requirements : List[str]
        The requirements of the skill.
    created_at : str
        The date and time when the skill was created.
    updated_at : str
        The date and time when the skill was last updated.
    data : WaldiezSkillData
        The data of the skill. See `WaldiezSkillData`.
    """

    id: Annotated[
        str, Field(..., title="ID", description="The ID of the skill.")
    ]
    type: Annotated[
        Literal["skill"],
        Field(
            default="skill",
            title="Type",
            description="The type of the 'node' in a graph.",
        ),
    ]
    name: Annotated[
        str, Field(..., title="Name", description="The name of the skill.")
    ]
    description: Annotated[
        str,
        Field(
            ...,
            title="Description",
            description="The description of the skill.",
        ),
    ]
    tags: Annotated[
        List[str],
        Field(
            title="Tags",
            description="The tags of the skill.",
            default_factory=list,
        ),
    ]
    requirements: Annotated[
        List[str],
        Field(
            title="Requirements",
            description="The requirements of the skill.",
            default_factory=list,
        ),
    ]
    data: Annotated[
        WaldiezSkillData,
        Field(..., title="Data", description="The data of the skill."),
    ]
    created_at: Annotated[
        str,
        Field(
            default_factory=now,
            title="Created At",
            description="The date and time when the skill was created.",
        ),
    ]
    updated_at: Annotated[
        str,
        Field(
            default_factory=now,
            title="Updated At",
            description="The date and time when the skill was last updated.",
        ),
    ]

    @property
    def is_shared(self) -> bool:
        """Check if the skill is shared.

        Returns
        -------
        bool
            True if the skill is shared, False otherwise.
        """
        return self.name == SHARED_SKILL_NAME

    def get_content(self) -> str:
        """Get the content of the skill.

        Returns
        -------
        str
            The content of the skill.
        """
        if self.name == SHARED_SKILL_NAME:
            # the whole content (globals)
            return self.data.content
        return get_function(self.data.content, self.name)

    @model_validator(mode="after")
    def validate_data(self) -> Self:
        """Validate the data.

        Returns
        -------
        WaldiezSkill
            The skill.

        Raises
        ------
        ValueError
            If the skill name is not in the content.
            If the skill content is invalid.
        """
        search = f"def {self.name}("
        if self.name != SHARED_SKILL_NAME and search not in self.data.content:
            raise ValueError(
                f"The skill name '{self.name}' is not in the content."
            )
        error, tree = parse_code_string(self.data.content)
        if error is not None or tree is None:
            raise ValueError(f"Invalid skill content: {error}")
        return self

    @property
    def content(self) -> str:
        """Get the content (source) of the skill."""
        return self.data.content

    @property
    def secrets(self) -> Dict[str, str]:
        """Get the secrets (environment variables) of the skill."""
        return self.data.secrets or {}



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa: E501
"""Try to install pysqlite3-binary.

Highly recommended to be run in a virtual environment.
'setuptools' and 'wheel' will also be installed if not already installed.
"""

import os
import shutil
import site
import subprocess
import sys
import tempfile
import urllib.request
import zipfile

PYSQLITE3_VERSION = "0.5.4"
SQLITE_URL = "https://www.sqlite.org/2025/sqlite-amalgamation-3480000.zip"
PYSQLITE3_URL = f"https://github.com/coleifer/pysqlite3/archive/refs/tags/{PYSQLITE3_VERSION}.zip"  # pylint: disable=line-too-long

PIP = f"{sys.executable} -m pip"


def run_command(command: str, cwd: str = ".") -> None:
    """Run a command.

    Parameters
    ----------
    command : str
        The command to run.
    cwd : str
        The current working directory.
    """
    cmd_parts = command.split(" ")
    if cwd == ".":
        cwd = os.getcwd()
    try:
        subprocess.run(  # nosemgrep  # nosec
            cmd_parts,
            check=True,
            cwd=cwd,
            env=os.environ,
            encoding="utf-8",
        )
    except subprocess.CalledProcessError as e:
        print(f"Error running command: {e}")
        sys.exit(1)


def in_virtualenv() -> bool:
    """Check if the script is running in a virtual environment.

    Returns
    -------
    bool
        True if in a virtual environment, False otherwise.
    """
    return hasattr(sys, "real_prefix") or (
        hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
    )


def pip_install(*package_names: str, cwd: str = ".") -> None:
    """Install packages using pip.

    Parameters
    ----------
    *package_names : tuple[str, ...]
        The package names or paths to install.
    cwd : str
        The current working directory.
    """
    args = "-qq"
    if not in_virtualenv():
        args += " --user --break-system-packages"
    package_names_str = " ".join(package_names)
    run_command(f"{PIP} install {args} {package_names_str}", cwd)


def pip_uninstall(*package_names: str, cwd: str = ".") -> None:
    """Uninstall packages using pip.

    Parameters
    ----------
    *package_names : tuple[str, ...]
        The package names to uninstall.
    cwd : str
        The current working directory.
    """
    args = "-qq --yes"
    if not in_virtualenv():
        args += " --break-system-packages"
    package_names_str = " ".join(package_names)
    run_command(f"{PIP} uninstall {args} {package_names_str}", cwd)


def download_sqlite_amalgamation() -> str:
    """Download the SQLite amalgamation source code.

    Returns
    -------
    str
        The path to the extracted SQLite source code.
    """
    zip_path = "sqlite_amalgamation.zip"
    extract_path = "sqlite_amalgamation"

    # Download the SQLite source code
    print("Downloading SQLite amalgamation source code...")
    urllib.request.urlretrieve(SQLITE_URL, zip_path)  # nosec

    # Extract the SQLite source code
    print("Extracting SQLite source code...")
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(extract_path)

    # Clean up the zip file
    os.remove(zip_path)

    # Return the path to the extracted source code
    return os.path.join(extract_path, "sqlite-amalgamation-3480000")


def rename_package_name(pysqlite3_dir: str) -> None:
    """Rename the package name in the setup.py file.

    Parameters
    ----------
    pysqlite3_dir : str
        The path to the pysqlite3 directory.
    """
    setup_file = os.path.join(pysqlite3_dir, "setup.py")
    with open(setup_file, "r", encoding="utf-8") as file:
        setup_py = file.read()
    # sed -i "s|name='pysqlite3-binary'|name=PACKAGE_NAME|g" setup.py
    setup_py = setup_py.replace(
        "name='pysqlite3'", "name='pysqlite3-binary'"
    ).replace("name=PACKAGE_NAME,", "name='pysqlite3-binary',")
    with open(setup_file, "w", encoding="utf-8", newline="\n") as file:
        file.write(setup_py)


def prepare_pysqlite3(sqlite_amalgamation_path: str) -> str:
    """Prepare pysqlite3 using the SQLite amalgamation source code.

    Parameters
    ----------
    sqlite_amalgamation_path : str
        The path to the SQLite amalgamation source code.

    Returns
    -------
    str
        The path to the pysqlite3 directory.
    """
    pysqlite3_zip = "pysqlite3.zip"
    pysqlite3_extract = "pysqlite3"
    urllib.request.urlretrieve(PYSQLITE3_URL, pysqlite3_zip)  # nosec
    with zipfile.ZipFile(pysqlite3_zip, "r") as zip_ref:
        zip_ref.extractall(pysqlite3_extract)
    os.remove(pysqlite3_zip)
    sqlite3_c = os.path.join(sqlite_amalgamation_path, "sqlite3.c")
    sqlite3_h = os.path.join(sqlite_amalgamation_path, "sqlite3.h")
    pysqlite3_dir = os.path.join(
        pysqlite3_extract, f"pysqlite3-{PYSQLITE3_VERSION}"
    )
    shutil.copy(sqlite3_c, pysqlite3_dir)
    shutil.copy(sqlite3_h, pysqlite3_dir)
    rename_package_name(pysqlite3_dir)
    return pysqlite3_dir


def install_pysqlite3(sqlite_amalgamation_path: str) -> None:
    """Install pysqlite3 using the SQLite amalgamation source code.

    Parameters
    ----------
    sqlite_amalgamation_path : str
        The path to the SQLite amalgamation source code.
    """
    # pylint: disable=too-many-try-statements
    try:
        pysqlite3_dir = prepare_pysqlite3(sqlite_amalgamation_path)
        pip_install("setuptools")
        run_command(f"{sys.executable} setup.py build_static", pysqlite3_dir)
        pip_install("wheel")
        run_command(
            f"{PIP} wheel . -w dist",
            pysqlite3_dir,
        )
        wheel_file = os.listdir(os.path.join(pysqlite3_dir, "dist"))[0]
        wheel_path = os.path.join("dist", wheel_file)
        pip_install(wheel_path, cwd=pysqlite3_dir)
    except BaseException as e:  # pylint: disable=broad-except
        print(f"Failed to install pysqlite3: {e}")
        sys.exit(1)


def check_pysqlite3() -> None:
    """Check the installation of pysqlite3."""
    # pylint: disable=unused-import, import-outside-toplevel, line-too-long
    try:
        import pysqlite3  # type: ignore[unused-ignore, import-untyped, import-not-found]  # noqa
    except ImportError:
        print("pysqlite3 not found or cannot be imported.")
        # Uninstall pysqlite3-binary if it is already installed
        pip_uninstall("pysqlite3", "pysqlite3-binary")
        source_path = download_sqlite_amalgamation()
        install_pysqlite3(source_path)
        site.main()
        # Re-import pysqlite3 as sqlite3
        import pysqlite3  # type: ignore[unused-ignore, import-untyped, import-not-found]  # noqa


def test_sqlite_usage() -> None:
    """Test the usage of the sqlite3 module."""
    # pylint: disable=import-outside-toplevel
    sys.modules["sqlite3"] = sys.modules.pop("pysqlite3")
    import sqlite3  # noqa

    print(sqlite3.__file__)
    # it should be sth like: /path/to/site-packages/pysqlite3/__init__.py
    conn = sqlite3.connect(":memory:")
    cursor = conn.cursor()
    cursor.execute("CREATE TABLE test (id INTEGER PRIMARY KEY, name TEXT)")
    cursor.execute("INSERT INTO test (name) VALUES ('test')")
    cursor.execute("SELECT * FROM test")
    rows = cursor.fetchall()
    print(rows)
    conn.close()


def main() -> None:
    """Run the check."""
    if "--force" in sys.argv:
        pip_uninstall("pysqlite3", "pysqlite3-binary")
    cwd = os.getcwd()
    tmpdir = tempfile.mkdtemp()
    os.chdir(tmpdir)
    check_pysqlite3()
    os.chdir(cwd)
    shutil.rmtree(tmpdir)
    test_sqlite_usage()


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=line-too-long
"""Check for conflicts with 'autogen-agentchat' package."""

import sys
from importlib.metadata import PackageNotFoundError, version

__WALDIEZ_CHECKED_FOR_CONFLICTS = False


# fmt: off
def _check_conflicts() -> None:  # pragma: no cover
    """Check for conflicts with 'autogen-agentchat' package."""
    try:
        version("autogen-agentchat")
        print(
            "Conflict detected: 'autogen-agentchat' is installed "
            "in the current environment, \n"
            "which conflicts with 'ag2' / 'pyautogen'.\n"
            "Please uninstall 'autogen-agentchat': \n"
            f"{sys.executable} -m pip uninstall -y autogen-agentchat" + "\n"
            "And install 'pyautogen' (and/or 'waldiez') again: \n"
            f"{sys.executable} -m pip install --force pyautogen waldiez",
            file=sys.stderr,
        )
        sys.exit(1)
    except PackageNotFoundError:
        pass
# fmt: on


def check_conflicts() -> None:  # pragma: no cover
    """Check for conflicts with 'autogen-agentchat' package."""
    # pylint: disable=global-statement
    global __WALDIEZ_CHECKED_FOR_CONFLICTS
    if __WALDIEZ_CHECKED_FOR_CONFLICTS is False:
        _check_conflicts()
        __WALDIEZ_CHECKED_FOR_CONFLICTS = True



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Export reasoning agent to string."""

from typing import Callable

from waldiez.models import WaldiezAgent, WaldiezReasoningAgent


def get_reasoning_agent_extras(
    agent: WaldiezAgent,
    serializer: Callable[..., str],
) -> str:
    """Get the reasoning agent extras.

    Parameters
    ----------
    agent : WaldiezReasoningAgent
        The reasoning agent.
    serializer : Callable[..., str]
        The serializer to get the string representation of an object.

    Returns
    -------
    str
        The reasoning agent extras.
    """
    if agent.agent_type != "reasoning" or not isinstance(
        agent, WaldiezReasoningAgent
    ):
        return ""
    reasoning_config = agent.get_reasoning_config()
    serialized = serializer(reasoning_config)
    content = "\n    verbose=" + f"{agent.data.verbose},"
    content += "\n    reason_config=" + f"{serialized},"
    return content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Exporting teachability data for agents."""

from typing import Dict

from waldiez.models import WaldiezAgent


def get_agent_teachability_string(
    agent: WaldiezAgent,
    agent_names: Dict[str, str],
) -> str:
    """Get the teachability string to use for the agent.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    agent_names : Dict[str, str]
        A mapping of agent id to agent name.

    Returns
    -------
    str
        The teachability string
    """
    if not agent.data.teachability or not agent.data.teachability.enabled:
        return ""
    agent_name = agent_names[agent.id]
    teachability = agent.data.teachability
    content = f"{agent_name}_teachability = teachability.Teachability(" + "\n"
    content += f"    verbosity={teachability.verbosity}," + "\n"
    content += f"    reset_db={teachability.reset_db}," + "\n"
    content += f"    recall_threshold={teachability.recall_threshold}," + "\n"
    content += (
        f"    max_num_retrievals={teachability.max_num_retrievals}," + "\n"
    )
    content += ")\n\n\n"
    content += f"{agent_name}_teachability.add_to_agent({agent_name})"
    return content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Swarm Agent ON_CONDITION Available Model."""

from typing import Optional, Tuple

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from ...common import WaldiezBase, check_function, generate_function

WaldiezSwarmOnConditionAvailableCheckType = Literal[
    "string", "callable", "none"
]

CUSTOM_ON_CONDITION_AVAILABLE = "custom_on_condition_available"
CUSTOM_ON_CONDITION_AVAILABLE_ARGS = ["agent", "message"]
CUSTOM_ON_CONDITION_AVAILABLE_TYPES = (
    ["ConversableAgent", "Dict[str, Any]"],
    "bool",
)

# In ag2 it is used as:
#
# if on_condition.available is not None:
#     if isinstance(on_condition.available, Callable):
#         is_available = on_condition.available(
#           agent, next(iter(agent.chat_messages.values()))
#         )
#     elif isinstance(on_condition.available, str):
#         is_available = agent.get_context(on_condition.available) or False


class WaldiezSwarmOnConditionAvailable(WaldiezBase):
    """Swarm condition availability check."""

    type: Annotated[
        WaldiezSwarmOnConditionAvailableCheckType,
        Field(
            "none",
            alias="availableCheckType",
            title="Available Check Type",
            description=("The type of the `available` property to check. "),
        ),
    ] = "none"
    value: Annotated[
        Optional[str],
        Field(
            None,
            title="Available",
            description=(
                "Optional condition to determine if this ON_CONDITION "
                "is available. Can be a Callable or a string.  If a string, "
                " it will look up the value of the context variable with that "
                "name, which should be a bool."
            ),
        ),
    ]

    _available_string: str = ""

    @property
    def available_string(self) -> str:
        """Get the available string.

        Returns
        -------
        str
            The available string.
        """
        return self._available_string

    def get_available(
        self,
        name_prefix: Optional[str] = None,
        name_suffix: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Get the available string.

        Parameters
        ----------
        name_prefix : str, optional
            The prefix to add to the function name. Default is None.
        name_suffix : str, optional
            The suffix to add to the function name. Default is None.
        Returns
        -------
        Tuple[str, str]
            The available string or function name and code if available.
        """
        if self.type == "none" or not self.value:
            return "", ""
        if self.type == "string":
            return self._available_string, ""
        function_name = CUSTOM_ON_CONDITION_AVAILABLE
        if name_prefix:
            function_name = f"{name_prefix}_{function_name}"
        if name_suffix:
            function_name = f"{function_name}_{name_suffix}"
        return function_name, generate_function(
            function_name=function_name,
            function_args=CUSTOM_ON_CONDITION_AVAILABLE_ARGS,
            function_types=CUSTOM_ON_CONDITION_AVAILABLE_TYPES,
            function_body=self._available_string,
        )

    @model_validator(mode="after")
    def validate_available(self) -> Self:
        """Validate the available check.

        Returns
        -------
        Self
            The swarm agent's on condition available model.

        Raises
        ------
        ValueError
            If the available check fails.
        """
        if self.type == "callable":
            if not self.value:
                raise ValueError(
                    "A callable is expected, but no value was provided."
                )
            is_valid, error_or_body = check_function(
                code_string=self.value,
                function_name=CUSTOM_ON_CONDITION_AVAILABLE,
                function_args=CUSTOM_ON_CONDITION_AVAILABLE_ARGS,
            )
            if not is_valid:
                raise ValueError(error_or_body)
            self._available_string = error_or_body
        if self.type == "string":
            if not self.value:
                raise ValueError(
                    "A string is expected, but no value was provided."
                )
            self._available_string = self.value
        return self



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get the `is_termination_message` check for the agent."""

from typing import Tuple

from waldiez.models import WaldiezAgent


def get_is_termination_message(
    agent: WaldiezAgent, agent_name: str
) -> Tuple[str, str]:
    """Get the `is_termination_message` argument and content (if any).

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    agent_name : str
        The agent name.

    Returns
    -------
    Tuple[str, str]
        - The termination function name or lambda or None.
        - The termination function definition and content if any.

    Raises
    ------
    ValueError
        If the termination type is invalid.
    """
    if agent.data.termination.type == "none":
        return "None", ""
    if agent.data.termination.type == "keyword":
        return agent.data.termination.string, ""
    if agent.data.termination.type == "method":
        content, function_name = (
            agent.data.termination.get_termination_function(
                name_suffix=agent_name
            )
        )
        return function_name, "\n\n" + content + "\n"
    raise ValueError(f"Invalid termination type: {agent.data.termination.type}")



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez models.
- Agents (Users, Assistants, Group Managers, etc.).
- Chat (Messages, Summaries, etc.).
- Model (LLM config, API type, etc.).
- Skill (Skills/Tools to be registered).
- Flow (Flow of the conversation).
- Methods (Method names, arguments, hints, etc.).
- Waldiez (Main class to hold the flow).
"""

from .agents import (
    WaldiezAgent,
    WaldiezAgentCodeExecutionConfig,
    WaldiezAgentData,
    WaldiezAgentLinkedSkill,
    WaldiezAgentNestedChat,
    WaldiezAgentNestedChatMessage,
    WaldiezAgents,
    WaldiezAgentTeachability,
    WaldiezAgentTerminationMessage,
    WaldiezAgentType,
    WaldiezAssistant,
    WaldiezAssistantData,
    WaldiezGroupManager,
    WaldiezGroupManagerData,
    WaldiezGroupManagerSpeakers,
    WaldiezGroupManagerSpeakersSelectionMethod,
    WaldiezGroupManagerSpeakersSelectionMode,
    WaldiezGroupManagerSpeakersTransitionsType,
    WaldiezRagUser,
    WaldiezRagUserChunkMode,
    WaldiezRagUserData,
    WaldiezRagUserModels,
    WaldiezRagUserRetrieveConfig,
    WaldiezRagUserTask,
    WaldiezRagUserVectorDb,
    WaldiezRagUserVectorDbConfig,
    WaldiezReasoningAgent,
    WaldiezReasoningAgentData,
    WaldiezReasoningAgentReasonConfig,
    WaldiezSwarmAfterWork,
    WaldiezSwarmAfterWorkOption,
    WaldiezSwarmAfterWorkRecipientType,
    WaldiezSwarmAgent,
    WaldiezSwarmAgentData,
    WaldiezSwarmHandoff,
    WaldiezSwarmOnCondition,
    WaldiezSwarmOnConditionAvailable,
    WaldiezSwarmOnConditionTarget,
    WaldiezSwarmUpdateSystemMessage,
    WaldiezUserProxy,
    WaldiezUserProxyData,
)
from .chat import (
    WaldiezChat,
    WaldiezChatData,
    WaldiezChatMessage,
    WaldiezChatNested,
    WaldiezChatSummary,
    WaldiezChatSummaryMethod,
)
from .flow import WaldiezFlow, WaldiezFlowData
from .model import (
    WaldiezModel,
    WaldiezModelAPIType,
    WaldiezModelData,
    WaldiezModelPrice,
)
from .skill import SHARED_SKILL_NAME, WaldiezSkill, WaldiezSkillData
from .waldiez import Waldiez

# pylint: disable=duplicate-code
__all__ = [
    "SHARED_SKILL_NAME",
    "Waldiez",
    "WaldiezAgent",
    "WaldiezAgentCodeExecutionConfig",
    "WaldiezAgentData",
    "WaldiezAgentLinkedSkill",
    "WaldiezAgentNestedChat",
    "WaldiezAgentNestedChatMessage",
    "WaldiezAgents",
    "WaldiezAgentTeachability",
    "WaldiezAgentTerminationMessage",
    "WaldiezAgentType",
    "WaldiezAssistant",
    "WaldiezAssistantData",
    "WaldiezChat",
    "WaldiezChatData",
    "WaldiezChatSummary",
    "WaldiezChatNested",
    "WaldiezChatSummaryMethod",
    "WaldiezFlow",
    "WaldiezFlowData",
    "WaldiezGroupManager",
    "WaldiezGroupManagerData",
    "WaldiezGroupManagerSpeakers",
    "WaldiezGroupManagerSpeakersSelectionMethod",
    "WaldiezGroupManagerSpeakersSelectionMode",
    "WaldiezGroupManagerSpeakersTransitionsType",
    "WaldiezChatMessage",
    "WaldiezModel",
    "WaldiezModelAPIType",
    "WaldiezModelData",
    "WaldiezModelPrice",
    "WaldiezRagUser",
    "WaldiezRagUserData",
    "WaldiezReasoningAgent",
    "WaldiezReasoningAgentData",
    "WaldiezReasoningAgentReasonConfig",
    "WaldiezSkill",
    "WaldiezSkillData",
    "WaldiezUserProxy",
    "WaldiezUserProxyData",
    "WaldiezRagUserRetrieveConfig",
    "WaldiezRagUserTask",
    "WaldiezRagUserChunkMode",
    "WaldiezRagUserVectorDb",
    "WaldiezRagUserVectorDbConfig",
    "WaldiezRagUserModels",
    "WaldiezSwarmAgent",
    "WaldiezSwarmAfterWork",
    "WaldiezSwarmAfterWorkOption",
    "WaldiezSwarmAfterWorkRecipientType",
    "WaldiezSwarmAgentData",
    "WaldiezSwarmHandoff",
    "WaldiezSwarmOnCondition",
    "WaldiezSwarmOnConditionTarget",
    "WaldiezSwarmOnConditionAvailable",
    "WaldiezSwarmUpdateSystemMessage",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# https://docs.ag2.ai/docs/reference/agentchat/contrib/swarm_agent
"""Swarm agent data."""

from typing import List, Union

from pydantic import Field, model_validator
from typing_extensions import Annotated, Self

from ..agent import WaldiezAgentData
from .after_work import WaldiezSwarmAfterWork
from .on_condition import WaldiezSwarmOnCondition
from .update_system_message import WaldiezSwarmUpdateSystemMessage

WaldiezSwarmHandoff = Union[WaldiezSwarmOnCondition, WaldiezSwarmAfterWork]


# flake8: noqa: E501
# pylint: disable=line-too-long
class WaldiezSwarmAgentData(WaldiezAgentData):
    """Swarm agent data.

    Attributes
    ----------
    is_initial: bool
        Whether the agent is the initial agent.
    functions : List[str]
        A list of functions (skill ids) to register with the agent.

    update_agent_state_before_reply : List[str]
        A list of functions, including `UPDATE_SYSTEM_MESSAGE`,
        called to update the agent's state before it replies. Each function
        is called when the agent is selected and before it speaks.

    handoffs : List[Union[WaldiezSwarmOnCondition, WaldiezSwarmAfterWork]]
        A list of hand offs to register.

    Notes
    -----
    Each agent should have at most one `AfterWork` and (if any) it should be
    at the end the list of hand offs.
    """

    is_initial: Annotated[
        bool,
        Field(
            False,
            title="Is Initial",
            alias="isInitial",
            description=("Whether the agent is the initial agent."),
        ),
    ]
    functions: Annotated[
        List[str],
        Field(
            title="Functions",
            description=(
                "A list of functions (skill ids) to register with the agent."
            ),
            default_factory=list,
        ),
    ]
    update_agent_state_before_reply: Annotated[
        List[Union[str, WaldiezSwarmUpdateSystemMessage]],
        Field(
            title="Update Agent State Before Reply",
            alias="updateAgentStateBeforeReply",
            description=(
                "A list of functions, including UPDATE_SYSTEM_MESSAGEs,"
                "called to update the agent's state before it replies. "
                " Each function is called when the agent is selected "
                "and before it speaks. If not an UPDATE_SYSTEM_MESSAGE, "
                "it should be a skill id."
            ),
            default_factory=list,
        ),
    ]
    handoffs: Annotated[
        List[Union[WaldiezSwarmOnCondition, WaldiezSwarmAfterWork]],
        Field(
            title="Handoffs",
            description=(
                "A list of hand offs to register. "
                "There should only be at most one `AfterWork` per agent"
                "And (if any) it should be at the end of the list."
            ),
            default_factory=list,
        ),
    ]

    @model_validator(mode="after")
    def validate_handoffs(self) -> Self:
        """Validate the hand offs.

        Returns
        -------
        Self
            The swarm agent data.

        Raises
        ------
        ValueError
            If there are more than one `AfterWork`s.
        """
        after_works: List[WaldiezSwarmAfterWork] = [
            hand_off
            for hand_off in self.handoffs
            if isinstance(hand_off, WaldiezSwarmAfterWork)
        ]
        if len(after_works) > 1:
            raise ValueError(
                "Each agent should have at most one `AfterWork` "
                "and (if any) it should be at the end of the list."
            )
        on_conditions: List[WaldiezSwarmOnCondition] = [
            hand_off
            for hand_off in self.handoffs
            if isinstance(hand_off, WaldiezSwarmOnCondition)
        ]
        on_conditions = sorted(on_conditions, key=lambda x: x.target.order)
        handoffs = on_conditions + after_works
        self.handoffs = handoffs
        return self



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""User proxy agent model."""

from .user_proxy import WaldiezUserProxy
from .user_proxy_data import WaldiezUserProxyData

__all__ = ["WaldiezUserProxy", "WaldiezUserProxyData"]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Group chat manager agent."""

from typing import List, Literal

from pydantic import Field
from typing_extensions import Annotated

from ..agent import WaldiezAgent
from .group_manager_data import WaldiezGroupManagerData
from .speakers import WaldiezGroupManagerSpeakers


class WaldiezGroupManager(WaldiezAgent):
    """Group chat manager agent.

    A `WaldiezAgent` with agent_type `manager`, `human_input_mode`: `"NEVER"`
    and chat group related config for the agent.
    Also see `WaldiezAgent`, `WaldiezGroupManagerData`, `WaldiezAgentData`

    Attributes
    ----------
    agent_type : Literal["manager"]
        The agent type: 'manager' for a group manager agent
    data : WaldiezGroupManagerData
        The group manager agent's data.

    Functions
    ---------
    validate_transitions(agent_ids: List[str])
        Validate the transitions.
    """

    agent_type: Annotated[
        Literal["manager"],
        Field(
            "manager",
            title="Agent type",
            description="The agent type: 'manager' for a group manager agent",
            alias="agentType",
        ),
    ]
    data: Annotated[
        WaldiezGroupManagerData,
        Field(
            title="Data",
            description="The group manager agent's data",
            default_factory=WaldiezGroupManagerData,
        ),
    ]

    def validate_transitions(self, agent_ids: List[str]) -> None:
        """Validate the transitions.

        If the selection mode is `transition`:

        - if `allow_repeat` is a list of agent_ids,
                make sure these ids exist.
        - make sure the `allowed_or_disallowed_transitions` mapping
                has valid agent ids.

        Parameters
        ----------
        agent_ids : List[str]
            The list of agent IDs.

        Raises
        ------
        ValueError
            If the transitions are invalid.
        """
        speakers: WaldiezGroupManagerSpeakers = self.data.speakers
        if speakers.selection_mode != "transition":
            return
        allow_repeat = speakers.allow_repeat
        if isinstance(allow_repeat, list):
            for agent_id in allow_repeat:
                if agent_id not in agent_ids:
                    raise ValueError(f"Invalid agent id: {agent_id}")
        for (
            agent_id,
            transitions,
        ) in speakers.allowed_or_disallowed_transitions.items():
            if agent_id not in agent_ids:
                raise ValueError(f"Invalid agent id: {agent_id}")
            for agent_id in transitions:
                if agent_id not in agent_ids:
                    raise ValueError(f"Invalid agent id: {agent_id}")



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""RAG User related exporting utils."""

from typing import Callable, Dict, List, Set, Tuple, Union

from waldiez.models import (
    WaldiezAgent,
    WaldiezRagUser,
    WaldiezRagUserModels,
    WaldiezRagUserRetrieveConfig,
)

from .vector_db import get_rag_user_vector_db_string


def get_rag_user_extras(
    agent: WaldiezAgent,
    agent_name: str,
    model_names: Dict[str, str],
    path_resolver: Callable[[str], str],
    serializer: Callable[..., str],
) -> Tuple[str, str, Set[str]]:
    """Get the RAG user extra argument, imports and content before the agent.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    agent_name : str
        The agent's name.
    model_names : Dict[str, str]
        A mapping from model id to model name.
    path_resolver : Callable[[str], str]
        The path resolver function.
    serializer : Callable[..., str]
        The serializer function.

    Returns
    -------
    Tuple[str, str, Set[str]]
        The content before the agent, the retrieve arg and the db imports.
    """
    before_agent_string = ""
    retrieve_arg = ""
    db_imports: Set[str] = set()
    if agent.agent_type == "rag_user" and isinstance(agent, WaldiezRagUser):
        rag_content_before_agent, retrieve_arg, db_imports = (
            get_rag_user_retrieve_config_str(
                agent=agent,
                agent_name=agent_name,
                model_names=model_names,
                path_resolver=path_resolver,
                serializer=serializer,
            )
        )
        if retrieve_arg:
            retrieve_arg = "\n" + f"    retrieve_config={retrieve_arg},"
        if rag_content_before_agent:
            before_agent_string += rag_content_before_agent
    return before_agent_string, retrieve_arg, db_imports


# pylint: disable=too-many-locals
def get_rag_user_retrieve_config_str(
    agent: WaldiezRagUser,
    agent_name: str,
    model_names: Dict[str, str],
    path_resolver: Callable[[str], str],
    serializer: Callable[..., str],
) -> Tuple[str, str, Set[str]]:
    """Get the RAG user retrieve config string.

    Parameters
    ----------
    agent : WaldiezRagUser
        The agent.
    agent_name : str
        The agent's name.
    model_names : Dict[str, str]
        A mapping from model id to model name.
    path_resolver : Callable[[str], str]
        The path resolver function.
    serializer : Callable[..., str]
        The serializer function.
    Returns
    -------
    Tuple[str, str, Set[str]]
        The content before the args, the args and the imports.
    """
    # e.g. user_agent = RetrieveUserProxyAgent(
    # ...other common/agent args,
    #  retrieve_config={what_this_returns})
    imports: Set[str] = set()
    retrieve_config = agent.retrieve_config
    before_the_args, vector_db_arg, db_imports = get_rag_user_vector_db_string(
        agent=agent,
        agent_name=agent_name,
    )
    imports.update(db_imports)
    args_dict = _get_args_dict(
        agent, retrieve_config, model_names, path_resolver
    )
    if retrieve_config.use_custom_token_count:
        function_content, token_count_arg_name = (
            retrieve_config.get_custom_token_count_function(
                name_suffix=agent_name
            )
        )
        before_the_args += "\n" + function_content + "\n"
        args_dict["custom_token_count_function"] = token_count_arg_name
    if retrieve_config.use_custom_text_split:
        function_content, text_split_arg_name = (
            retrieve_config.get_custom_text_split_function(
                name_suffix=agent_name
            )
        )
        before_the_args += "\n" + function_content + "\n"
        args_dict["custom_text_split_function"] = text_split_arg_name
    # docs_path = args_dict.pop("docs_path", [])
    args_content = serializer(args_dict)
    # get the last line (where the dict ends)
    args_parts = args_content.split("\n")
    before_vector_db = args_parts[:-1]
    closing_arg = args_parts[-1]
    args_content = "\n".join(before_vector_db)
    # add the vector_db arg
    args_content += ",\n" + f'        "vector_db": {vector_db_arg},' + "\n"
    # we should not need to include the client, but let's do it
    # to avoid later issues (with telemetry or other client settings)
    # https://github.com/ag2ai/ag2/blob/main/autogen/agentchat/\
    #   contrib/retrieve_user_proxy_agent.py#L265-L266
    if (
        f"{agent_name}_client" in before_the_args
        and agent.retrieve_config.vector_db == "chroma"
    ):
        args_content += f'        "client": {agent_name}_client,' + "\n"
    args_content += closing_arg
    return before_the_args, args_content, imports


def _get_model_arg(
    agent: WaldiezRagUser,
    retrieve_config: WaldiezRagUserRetrieveConfig,
    model_names: Dict[str, str],
) -> str:  # pragma: no cover
    agent_models = agent.data.model_ids
    if agent_models:
        first_model = agent_models[0]
        first_model_name = model_names[first_model]
        new_model_name = f"{first_model_name}"
        return f"{new_model_name}"
    if retrieve_config.model in model_names:
        selected_model = model_names[retrieve_config.model]
        new_model_name = f"{selected_model}"
        return f"{new_model_name}"
    return WaldiezRagUserModels[retrieve_config.vector_db]


def _get_args_dict(
    agent: WaldiezRagUser,
    retrieve_config: WaldiezRagUserRetrieveConfig,
    model_names: Dict[str, str],
    path_resolver: Callable[[str], str],
) -> Dict[str, Union[str, List[str]]]:
    model_arg = _get_model_arg(agent, retrieve_config, model_names)
    args_dict: Dict[str, Union[str, List[str]]] = {
        "task": retrieve_config.task,
        "model": model_arg,
    }
    optional_args = [
        "chunk_token_size",
        "context_max_tokens",
        "customized_prompt",
        "customized_answer_prefix",
    ]
    for arg in optional_args:
        arg_value = getattr(retrieve_config, arg)
        if arg_value is not None:
            args_dict[arg] = arg_value
            args_dict[arg] = getattr(retrieve_config, arg)
    docs_path: Union[str, List[str]] = []
    if retrieve_config.docs_path:
        doc_paths = (
            retrieve_config.docs_path
            if isinstance(retrieve_config.docs_path, list)
            else [retrieve_config.docs_path]
        )
        docs_path = [
            item for item in [path_resolver(path) for path in doc_paths] if item
        ]
        args_dict["docs_path"] = docs_path
    if docs_path:
        args_dict["docs_path"] = docs_path
    non_optional_args = [
        "new_docs",
        "update_context",
        "get_or_create",
        "overwrite",
        "recursive",
        "chunk_mode",
        "must_break_at_empty_line",
        "collection_name",
        "distance_threshold",
    ]
    for arg in non_optional_args:
        args_dict[arg] = getattr(retrieve_config, arg)
    return args_dict



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa: E501
"""Logging related string generation functions.

Functions
---------
get_logging_start_string
    Get the string to start logging.
get_logging_stop_string
    Get the string to stop logging.
get_sqlite_to_csv_string
    Get the sqlite to csv conversion code string.
get_sqlite_to_csv_call_string
    Get the string to call the sqlite to csv conversion.
"""


def get_start_logging(tabs: int = 0) -> str:
    """Get the logging start call string.

    Parameters
    ----------
    tabs : int, optional
        The number of tabs to use for indentation, by default 0

    Returns
    -------
    str
        The logging start string.

    Example
    -------
    ```python
    >>> get_start_logging()
    def start_logging() -> None:
        \"\"\"Start logging.\"\"\"
        runtime_logging.start(
            logger_type="sqlite",
            config={"dbname": "flow.db"},
        )
    """
    tab = "    " * tabs
    content = f'''
{tab}def start_logging() -> None:
{tab}    """Start logging."""
{tab}    runtime_logging.start(
{tab}        logger_type="sqlite",
{tab}        config={{"dbname": "flow.db"}},
{tab}    )
'''
    return content


# pylint: disable=differing-param-doc,differing-type-doc
def get_sync_sqlite_out() -> str:
    """Get the sqlite to csv and json conversion code string.

    Returns
    -------
    str
        The sqlite to csv and json conversion code string.

    Example
    -------
    ```python
    >>> get_sqlite_outputs()
    def get_sqlite_out(dbname: str, table: str, csv_file: str) -> None:
        \"\"\"Convert a sqlite table to csv and json files.

        Parameters
        ----------
        dbname : str
            The sqlite database name.
        table : str
            The table name.
        csv_file : str
            The csv file name.
        \"\"\"
        conn = sqlite3.connect(dbname)
        query = f"SELECT * FROM {table}"  # nosec
        cursor = conn.execute(query)
        rows = cursor.fetchall()
        column_names = [description[0] for description in cursor.description]
        data = [dict(zip(column_names, row)) for row in rows]
        conn.close()
        with open(csv_file, "w", newline="", encoding="utf-8") as file:
            csv_writer = csv.DictWriter(file, fieldnames=column_names)
            csv_writer.writeheader()
            csv_writer.writerows(data)
        json_file = csv_file.replace(".csv", ".json")
        with open(json_file, "w", encoding="utf-8") as file:
            json.dump(data, file, indent=4, ensure_ascii=False)
    ```
    """
    content = "\n\n"
    content += (
        "def get_sqlite_out(dbname: str, table: str, csv_file: str) -> None:\n"
    )
    content += '    """Convert a sqlite table to csv and json files.\n\n'
    content += "    Parameters\n"
    content += "    ----------\n"
    content += "    dbname : str\n"
    content += "        The sqlite database name.\n"
    content += "    table : str\n"
    content += "        The table name.\n"
    content += "    csv_file : str\n"
    content += "        The csv file name.\n"
    content += '    """\n'
    content += "    conn = sqlite3.connect(dbname)\n"
    content += '    query = f"SELECT * FROM {table}"  # nosec\n'
    content += "    try:\n"
    content += "        cursor = conn.execute(query)\n"
    content += "    except sqlite3.OperationalError:\n"
    content += "        conn.close()\n"
    content += "        return\n"
    content += "    rows = cursor.fetchall()\n"
    content += "    column_names = [description[0] for description "
    content += "in cursor.description]\n"
    content += "    data = [dict(zip(column_names, row)) for row in rows]\n"
    content += "    conn.close()\n"
    content += (
        '    with open(csv_file, "w", newline="", encoding="utf-8") as file:\n'
    )
    content += (
        "        csv_writer = csv.DictWriter(file, fieldnames=column_names)\n"
    )
    content += "        csv_writer.writeheader()\n"
    content += "        csv_writer.writerows(data)\n"
    content += '    json_file = csv_file.replace(".csv", ".json")\n'
    content += '    with open(json_file, "w", encoding="utf-8") as file:\n'
    content += "        json.dump(data, file, indent=4, ensure_ascii=False)\n"
    content += "\n\n"
    return content


# pylint: disable=differing-param-doc,differing-type-doc,line-too-long
def get_async_sqlite_out() -> str:
    """Get the sqlite to csv and json conversion code string.

    Returns
    -------
    str
        The sqlite to csv and json conversion code string.

    Example
    -------
    ```python
    >>> get_sqlite_outputs()
    async def get_sqlite_out(dbname: str, table: str, csv_file: str) -> None:
        \"\"\"Convert a sqlite table to csv and json files.

        Parameters
        ----------
        dbname : str
            The sqlite database name.
        table : str
            The table name.
        csv_file : str
            The csv file name.
        \"\"\"
        conn = await aiosqlite.connect(dbname)
        query = f"SELECT * FROM {table}"  # nosec
        try:
            cursor = await conn.execute(query)
        except BaseException:  # pylint: disable=broad-except
            await conn.close()
            return
        rows = await cursor.fetchall()
        column_names = [description[0] for description in cursor.description]
        data = [dict(zip(column_names, row)) for row in rows]
        await cursor.close()
        await conn.close()
        async with aiofiles.open(csv_file, "w", newline="", encoding="utf-8") as file:
            csv_writer = csv.DictWriter(file, fieldnames=column_names)
            csv_writer.writeheader()
            csv_writer.writerows(data)
        json_file = csv_file.replace(".csv", ".json")
        async with aiofiles.open(json_file, "w", encoding="utf-8") as file:
            await file.write(json.dumps(data, indent=4, ensure_ascii=False)
    ```
    """
    content = "\n\n"
    content += "async def get_sqlite_out(dbname: str, table: str, csv_file: str) -> None:\n"
    content += '    """Convert a sqlite table to csv and json files.\n\n'
    content += "    Parameters\n"
    content += "    ----------\n"
    content += "    dbname : str\n"
    content += "        The sqlite database name.\n"
    content += "    table : str\n"
    content += "        The table name.\n"
    content += "    csv_file : str\n"
    content += "        The csv file name.\n"
    content += '    """\n'
    content += "    conn = await aiosqlite.connect(dbname)\n"
    content += '    query = f"SELECT * FROM {table}"  # nosec\n'
    content += "    try:\n"
    content += "        cursor = await conn.execute(query)\n"
    content += "    except BaseException:  # pylint: disable=broad-except\n"
    content += "        await conn.close()\n"
    content += "        return\n"
    content += "    rows = await cursor.fetchall()\n"
    content += "    column_names = [description[0] for description "
    content += "in cursor.description]\n"
    content += "    data = [dict(zip(column_names, row)) for row in rows]\n"
    content += "    await cursor.close()\n"
    content += "    await conn.close()\n"
    content += (
        '    async with aiofiles.open(csv_file, "w", newline="", '
        'encoding="utf-8") as file:\n'
    )
    content += '        csv_writer = AsyncDictWriter(file, fieldnames=column_names, dialect="unix")\n'
    content += "        await csv_writer.writeheader()\n"
    content += "        await csv_writer.writerows(data)\n"
    content += '    json_file = csv_file.replace(".csv", ".json")\n'
    content += '    async with aiofiles.open(json_file, "w", encoding="utf-8") as file:\n'
    content += "        await file.write(json.dumps(data, indent=4, ensure_ascii=False))\n"
    content += "\n\n"
    return content


def get_sqlite_out(is_async: bool) -> str:
    """Get the sqlite to csv and json conversion code string.

    Parameters
    ----------
    is_async : bool
        Whether to use async mode.

    Returns
    -------
    str
        The sqlite to csv and json conversion code string.
    """
    if is_async:
        return get_async_sqlite_out()
    return get_sync_sqlite_out()


def get_sqlite_out_call(tabs: int, is_async: bool) -> str:
    """Get the sqlite to csv and json conversion call string.

    Parameters
    ----------
    tabs : int
        The number of tabs to use for indentation
    is_async : bool
        Whether to use async mode

    Returns
    -------
    str
        The sqlite to csv conversion call string.

    Example
    -------
    ```python
    >>> get_sqlite_out_call()
    if not os.path.exists("logs"):
        os.makedirs("logs")
    for table in [
        "chat_completions",
        "agents",
        "oai_wrappers",
        "oai_clients",
        "version",
        "events",
        "function_calls",
    ]:
        dest = os.path.join("logs", f"{table}.csv")
        get_sqlite_out("flow.db", table, dest)
    ```
    """
    table_names = [
        "chat_completions",
        "agents",
        "oai_wrappers",
        "oai_clients",
        "version",
        "events",
        "function_calls",
    ]
    tab = "    " * tabs
    content = ""
    content += tab + 'if not os.path.exists("logs"):\n'
    content += tab + '    os.makedirs("logs")\n'
    content += tab + "for table in [\n"
    for table in table_names:
        content += tab + f'    "{table}",' + "\n"
    content += tab + "]:\n"
    content += tab + '    dest = os.path.join("logs", f"{table}.csv")' + "\n"
    if is_async:
        content += tab + '    await get_sqlite_out("flow.db", table, dest)\n'
    else:
        content += tab + '    get_sqlite_out("flow.db", table, dest)\n'
    return content


def get_stop_logging(tabs: int, is_async: bool) -> str:
    """Get the function to stop logging and gather logs.

    Parameters
    ----------
    tabs : int
        The number of tabs to use for indentation
    is_async : bool
        Whether to use async mode

    Returns
    -------
    str
        The logging stop string.

    Example
    -------
    ```python
    >>> get_logging_stop_string()
    def stop_logging() -> None:
        \"\"\"Stop logging.\"\"\"
        runtime_logging.stop()
        for table in [
            "chat_completions",
            "agents",
            "oai_wrappers",
            "oai_clients",
            "version",
            "events",
            "function_calls",
        ]:
            dest = os.path.join("logs", f"{table}.csv")
            get_sqlite_out("flow.db", table, dest)
    """
    tab = "    " * tabs
    content = "\n" + tab
    if is_async:
        content += "async "
    content += "def stop_logging() -> None:\n"
    content += '    """Stop logging."""\n'
    content += f"{tab}    runtime_logging.stop()\n"
    content += get_sqlite_out_call(tabs + 1, is_async)
    return content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""RAG user agent.

It extends a user agent and has RAG related parameters (`retrieve_config`).
"""

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..agent import WaldiezAgent
from .rag_user_data import WaldiezRagUserData
from .retrieve_config import WaldiezRagUserRetrieveConfig


class WaldiezRagUser(WaldiezAgent):
    """RAG user agent.

    It extends a user agent and has RAG related parameters.

    Attributes
    ----------
    agent_type : Literal["rag_user"]
        The agent type: 'rag_user' for a RAG user agent.
    data : WaldiezRagUserData
        The RAG user agent's data.
        See `WaldiezRagUserData` for more info.
    retrieve_config : WaldiezRagUserRetrieveConfig
        The RAG user agent's retrieve config.
    """

    agent_type: Annotated[
        Literal["rag_user"],
        Field(
            "rag_user",
            title="Agent type",
            description="The agent type: 'rag_user' for a RAG user agent",
            alias="agentType",
        ),
    ]

    data: Annotated[
        WaldiezRagUserData,
        Field(
            title="Data",
            description="The RAG user agent's data",
            default_factory=WaldiezRagUserData,
        ),
    ]

    @property
    def retrieve_config(self) -> WaldiezRagUserRetrieveConfig:
        """Get the retrieve config.

        Returns
        -------
        WaldiezRagUserRetrieveConfig
            The RAG user agent's retrieve config.
        """
        return self.data.retrieve_config



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa E501
# pylint: disable=line-too-long
"""Utils to generate the content of a flow."""

from typing import Callable, Dict, List, Optional

from waldiez.models import Waldiez


def get_py_content_start(waldiez: Waldiez) -> str:
    """Get the first part of the python script.

    Parameters
    ----------
    waldiez : Waldiez
        The waldiez object.
    Returns
    -------
    str
        The first part of the python script.
    """
    content = "#!/usr/bin/env python\n"
    content += "# flake8: noqa E501\n"
    content += get_pylint_ignore_comment(False)
    content += "# cspell: disable\n"
    content += f'"""{waldiez.name}.' + "\n\n"
    content += f"{waldiez.description}" + "\n\n"
    tags = ", ".join(waldiez.tags)
    content += f"Tags: {tags}" + "\n\n"
    requirements = " ".join(waldiez.requirements)
    content += f"Requirements: {requirements}" + "\n\n"
    content += '"""\n\n'
    return content


def get_ipynb_content_start(
    waldiez: Waldiez, comment: Callable[[bool, int], str]
) -> str:
    """Get the first part of the ipynb file.

    Parameters
    ----------
    waldiez : Waldiez
        The waldiez object.
    comment : Callable[[bool, int], str]
        The function to create a comment.
    Returns
    -------
    str
        The first part of the ipynb file.
    """
    content = f"{comment(True, 1)}{waldiez.name}." + "\n\n"
    content += f"{comment(True, 2)}{waldiez.description}" + "\n\n"
    content += f"{comment(True, 2)}Dependencies" + "\n\n"
    content += "import sys\n"
    requirements = " ".join(waldiez.requirements)
    if requirements:
        # fmt: off
        content += "# " + f"!{{sys.executable}} -m pip install -q {requirements}" + "\n"
        # fmt: on
    content += "# flake8: noqa E501"
    content += get_pylint_ignore_comment(True)
    content += "# cspell: disable\n"
    return content


PYLINT_RULES = [
    "line-too-long",
    "unknown-option-value",
    "unused-argument",
    "unused-import",
    "invalid-name",
    "import-error",
    "inconsistent-quotes",
    "missing-function-docstring",
    "missing-param-doc",
    "missing-return-doc",
]


def get_pylint_ignore_comment(
    notebook: bool, rules: Optional[List[str]] = None
) -> str:
    """Get the pylint ignore comment string.

    Parameters
    ----------
    notebook : bool
        Whether the comment is for a notebook.
    rules : Optional[List[str]], optional
        The pylint rules to ignore, by default None.

    Returns
    -------
    str
        The pylint ignore comment string.

    Example
    -------
    ```python
    >>> get_pylint_ignore_comment(True, ["invalid-name", "line-too-long"])

    # pylint: disable=invalid-name, line-too-long
    ```
    """
    if not rules:
        rules = PYLINT_RULES
    line = "# pylint: disable=" + ",".join(rules)
    if notebook is True:
        line = "\n" + line
    return line + "\n"


def get_after_run_content(
    waldiez: Waldiez,
    agent_names: Dict[str, str],
    tabs: int,
) -> str:
    """Get content to add after the flow is run.

    Parameters
    ----------
    waldiez : Waldiez
        The waldiez object.
    agent_names : Dict[str, str]
        The dictionary of agent names and their corresponding ids
    tabs : int
        The number of tabs to add before the content.
    Returns
    -------
    str
        The content to add after the flow is run.
    """
    # if th eflow has reasoning agents, we add
    # visualize_tree(agent._root) for each agent
    content = ""
    space = "    " * tabs
    for agent in waldiez.agents:
        if agent.agent_type == "reasoning":
            agent_name = agent_names[agent.id]
            content += f"""
{space}# pylint: disable=broad-except,too-many-try-statements
{space}try:
{space}{space}visualize_tree({agent_name}._root)  # pylint: disable=protected-access
{space}{space}if os.path.exists("tree_of_thoughts.png"):
{space}{space}{space}new_name = "{agent_name}_tree_of_thoughts.png"
{space}{space}{space}os.rename("tree_of_thoughts.png", new_name)
{space}except BaseException:
{space}{space}pass
{space}# save the tree to json
{space}try:
{space}{space}data = {agent_name}._root.to_dict()  # pylint: disable=protected-access
{space}{space}with open("{agent_name}_reasoning_tree.json", "w", encoding="utf-8") as f:
{space}{space}{space}json.dump(data, f)
{space}except BaseException:
{space}{space}pass
"""
    return content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez chat summary options."""

from typing import Any, Dict, Optional

from pydantic import (
    Field,
    FieldSerializationInfo,
    field_serializer,
    field_validator,
)
from typing_extensions import Annotated, Literal

from ..common import WaldiezBase

WaldiezChatSummaryMethod = Literal[
    "reflectionWithLlm",
    "lastMsg",
    "reflection_with_llm",
    "last_msg",
]


class WaldiezChatSummary(WaldiezBase):
    """Llm summary method options.

    Attributes
    ----------
    method : Optional[WaldiezChatSummaryMethod]
        The method to use for the LLM summary. Defaults to "last_msg".
    prompt : str
        The prompt for the LLM summary method.
    args : Optional[Dict[str, Any]]
        The additional arguments for the LLM summary method, by default None.
    """

    method: Annotated[
        Optional[WaldiezChatSummaryMethod],
        Field(
            "last_msg",
            title="Method",
            description="The method to use for the LLM summary.",
        ),
    ]
    prompt: Annotated[
        str,
        Field(
            "",
            title="Prompt",
            description="The prompt for the LLM summary method.",
        ),
    ]
    args: Annotated[
        Dict[str, str],
        Field(
            title="Arguments",
            description="The additional arguments for the LLM summary method.",
            default_factory=dict,
        ),
    ]

    @field_validator("method", mode="before")
    @classmethod
    def validate_summary_method(
        cls, value: Optional[WaldiezChatSummaryMethod]
    ) -> Optional[WaldiezChatSummaryMethod]:
        """Validate the summary method.

        Parameters
        ----------
        value : Optional[WaldiezChatSummaryMethod]
            The passed WaldiezChatSummaryMethod

        Returns
        -------
        Optional[WaldiezChatSummaryMethod]
            The validated message summary method
        """
        if str(value).lower() == "none":
            return None
        if value == "lastMsg":
            return "last_msg"
        if value == "reflectionWithLlm":
            return "reflection_with_llm"
        return value

    @field_serializer("method")
    @classmethod
    def serialize_summary_method(
        cls, value: Any, info: FieldSerializationInfo
    ) -> Any:
        """Serialize summary method.

        Parameters
        ----------
        value : Any
            The value to serialize.
        info : FieldSerializationInfo
            The serialization info.

        Returns
        -------
        Any
            The serialized value.
        """
        if info.by_alias is True:
            if value == "reflection_with_llm":
                return "reflectionWithLlm"
            if value == "last_msg":
                return "lastMsg"
        return value



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utilities for chat content after exporting."""

from typing import List, Tuple, Union

from waldiez.exporting.base import (
    AgentPosition,
    ExportPosition,
    ExportPositions,
)


def add_before_chat_content(
    chat_content: str,
    before_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
) -> str:
    """Add the before chat content.

    Parameters
    ----------
    chat_content : str
        The chat content.
    before_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        The before export.

    Returns
    -------
    str
        The chat content with the before chat content.
    """
    new_content = str(chat_content)
    for content, position in before_export:
        if (
            isinstance(position, ExportPosition)
            and position.position == ExportPositions.CHATS
        ):
            new_content = content + "\n" + new_content
    if not new_content.endswith("\n"):
        new_content += "\n"
    return new_content


def add_after_chat_content(
    chat_content: str,
    after_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
) -> str:
    """Add the after chat content.

    Parameters
    ----------
    chat_content : str
        The chat content.
    after_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        The after export.

    Returns
    -------
    str
        The chat content with the after chat content.
    """
    new_content = str(chat_content)
    for content, position in after_export:
        if (
            isinstance(position, ExportPosition)
            and position.position == ExportPositions.CHATS
        ):
            new_content += content + "\n"
    if not new_content.endswith("\n"):
        new_content += "\n"
    return new_content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez

A Waldiez class contains all the information that is needed to generate
and run an autogen workflow. It has the model/LLM configurations, the agent
definitions and their optional additional skills to be used.
"""

import json
import warnings
from dataclasses import dataclass
from functools import cache
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from .agents import WaldiezAgent
from .chat import WaldiezChat
from .flow import WaldiezFlow
from .model import WaldiezModel
from .skill import WaldiezSkill


@dataclass(frozen=True, slots=True)
class Waldiez:
    """Waldiez data class.

    It contains all the information to generate and run an autogen workflow.
    """

    flow: WaldiezFlow

    @classmethod
    def from_dict(
        cls,
        data: Dict[str, Any],
        flow_id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        requirements: Optional[List[str]] = None,
    ) -> "Waldiez":
        """Create a Waldiez from dict.

        Parameters
        ----------
        data : Dict[str, Any]
            The data.
        flow_id : Optional[str], optional
            The flow id, by default None (retrieved from data or generated).
        name : Optional[str], optional
            The name, by default None (retrieved from data).
        description : Optional[str], optional
            The description, by default None (retrieved from data).
        tags : Optional[List[str]], optional
            The tags, by default None (retrieved from data).
        requirements : Optional[List[str]], optional
            The requirements, by default None (retrieved from data).

        Returns
        -------
        Waldiez
            The Waldiez.
        """
        flow = _get_flow(
            data,
            flow_id=flow_id,
            name=name,
            description=description,
            tags=tags,
            requirements=requirements,
        )
        return cls(flow=WaldiezFlow.model_validate(flow))

    @classmethod
    def load(
        cls,
        waldiez_file: Union[str, Path],
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        requirements: Optional[List[str]] = None,
    ) -> "Waldiez":
        """Load a Waldiez from a file.

        Parameters
        ----------
        waldiez_file : Union[str, Path]
            The Waldiez file.
        name : Optional[str], optional
            The name, by default None.
        description : Optional[str], optional
            The description, by default None.
        tags : Optional[List[str]], optional
            The tags, by default None.
        requirements : Optional[List[str]], optional
            The requirements, by default None.

        Returns
        -------
        Waldiez
            The Waldiez.

        Raises
        ------
        ValueError
            If the file is not found or invalid JSON.
        """
        data: Dict[str, Any] = {}
        if not Path(waldiez_file).exists():
            raise ValueError(f"File not found: {waldiez_file}")
        with open(waldiez_file, "r", encoding="utf-8") as file:
            try:
                data = json.load(file)
            except json.decoder.JSONDecodeError as error:
                raise ValueError(f"Invalid JSON: {waldiez_file}") from error
        return cls.from_dict(
            data,
            name=name,
            description=description,
            tags=tags,
            requirements=requirements,
        )

    def model_dump_json(
        self, by_alias: bool = True, indent: Optional[int] = None
    ) -> str:
        """Get the model dump json.

        We use `by_alias=True` by default to use the alias (toCamel).

        Parameters
        ----------
        by_alias : bool, optional
            Use alias (toCamel), by default True.
        indent : Optional[int], optional
            The indent, by default None.

        Returns
        -------
        str
            The model dump json.
        """
        return self.flow.model_dump_json(by_alias=by_alias, indent=indent)

    @property
    def has_rag_agents(self) -> bool:
        """Check if the flow has RAG agents."""
        return any(agent.agent_type == "rag_user" for agent in self.agents)

    @property
    def has_multimodal_agents(self) -> bool:
        """Check if the flow has multimodal agents."""
        return any(agent.data.is_multimodal for agent in self.agents)

    @property
    def chats(self) -> List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]:
        """Get the chats."""
        return self.flow.ordered_flow

    @property
    def agents(self) -> Iterator[WaldiezAgent]:
        """Get the agents.

        Yields
        -------
        WaldiezAgent
            The flow agents.
        """
        yield from self.flow.data.agents.members

    @property
    def skills(self) -> Iterator[WaldiezSkill]:
        """Get the flow skills.

        Yields
        -------
        WaldiezSkill
            The skills.
        """
        yield from self.flow.data.skills

    @property
    def models(self) -> Iterator[WaldiezModel]:
        """Get the models.

        Yields
        -------
        WaldiezModel
            The flow models.
        """
        yield from self.flow.data.models

    @property
    def name(self) -> str:
        """Get the flow name."""
        return self.flow.name or "Waldiez Flow"

    @property
    def description(self) -> str:
        """Get the flow description."""
        return self.flow.description or "Waldiez Flow description"

    @property
    def tags(self) -> List[str]:
        """Get the flow tags."""
        return self.flow.tags

    @property
    def is_async(self) -> bool:
        """Check if the flow is asynchronous."""
        return self.flow.is_async

    @property
    def requirements(self) -> List[str]:
        """Get the flow requirements."""
        autogen_version = _get_autogen_version()
        requirements_list = filter(
            lambda requirement: not (
                requirement.startswith("pyautogen")
                or requirement.startswith("ag2")
                or requirement.startswith("autogen")
            ),
            self.flow.requirements,
        )
        requirements = set(requirements_list)
        if self.has_rag_agents:
            requirements.add(f"pyautogen[retrievechat]=={autogen_version}")
        else:
            requirements.add(f"pyautogen=={autogen_version}")
        if self.has_multimodal_agents:
            requirements.add(f"pyautogen[lmm]=={autogen_version}")
        # ref: https://github.com/ag2ai/ag2/blob/main/pyproject.toml
        models_with_additional_requirements = [
            "together",
            "gemini",
            "mistral",
            "groq",
            "anthropic",
            "cohere",
            "bedrock",
        ]
        for model in self.models:
            if model.data.api_type == "google":
                requirements.add(f"pyautogen[gemini]=={autogen_version}")
                continue
            if model.data.api_type in models_with_additional_requirements:
                requirements.add(
                    f"pyautogen[{model.data.api_type}]=={autogen_version}"
                )
        return sorted(list(requirements))

    def get_flow_env_vars(self) -> List[Tuple[str, str]]:
        """Get the flow environment variables.

        Returns
        -------
        List[Tuple[str, str]]
            The environment variables for the flow.
        """
        env_vars: List[Tuple[str, str]] = []
        for skill in self.skills:
            for secret_key, secret_value in skill.secrets.items():
                env_vars.append((secret_key, secret_value))
        return env_vars

    def get_group_chat_members(self, agent: WaldiezAgent) -> List[WaldiezAgent]:
        """Get the chat members that connect to a group chat manager agent.

        Parameters
        ----------
        agent : WaldiezAgent
            The agent (group chat manager).

        Returns
        -------
        List[WaldiezAgent]
            The group chat members.
        """
        if agent.agent_type != "manager":
            return []
        return self.flow.get_group_chat_members(agent.id)

    def get_swarm_members(
        self, initial_agent: WaldiezAgent
    ) -> Tuple[List[WaldiezAgent], Optional[WaldiezAgent]]:
        """Get the chat members that connect to a swarm agent.

        Parameters
        ----------
        initial_agent : WaldiezAgent
            The initial agent.

        Returns
        -------
        Tuple[List[WaldiezAgent], Optional[WaldiezAgent]]
            The swarm agents and the user agent.
        """
        return self.flow.get_swarm_chat_members(initial_agent)


def _get_flow(
    data: Dict[str, Any],
    flow_id: Optional[str] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    tags: Optional[List[str]] = None,
    requirements: Optional[List[str]] = None,
) -> Dict[str, Any]:
    """Get the flow."""
    item_type = data.get("type", "flow")
    if item_type != "flow":
        # empty flow (from exported model/skill ?)
        raise ValueError(f"Invalid flow type: {item_type}")
    from_args = {
        "id": flow_id,
        "name": name,
        "description": description,
        "tags": tags,
        "requirements": requirements,
    }
    for key, value in from_args.items():
        if value:
            data[key] = value
    if "name" not in data:
        data["name"] = "Waldiez Flow"
    if "description" not in data:
        data["description"] = "Waldiez Flow description"
    if "tags" not in data:
        data["tags"] = []
    if "requirements" not in data:
        data["requirements"] = []
    return data


@cache
def _get_autogen_version() -> str:
    """Get the autogen version."""
    # pylint: disable=import-outside-toplevel
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            from autogen.version import __version__ as ag2  # type: ignore
        except ImportError as error:  # pragma: no cover
            raise ValueError("pyautogen is not installed.") from error
    return ag2



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Agent Termination Message Check."""

from typing import List, Optional, Tuple

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from ...common import WaldiezBase, check_function, generate_function

IS_TERMINATION_MESSAGE = "is_termination_message"
IS_TERMINATION_MESSAGE_ARGS = ["message"]
IS_TERMINATION_MESSAGE_TYPES = (
    ["Dict[str, Any]"],
    "bool",
)


class WaldiezAgentTerminationMessage(WaldiezBase):
    """Waldiez Agent Termination Message Check.

    Attributes
    ----------
    type : Literal["none", "keyword", "method"]
        The type of the termination check to use: "none", "keyword", "method"
    keywords : List[str]
        If the type is "keyword", the keywords to search in the message.
    criterion : Optional[Literal["found", "ending", "exact"]] = None
        If the type is "keyword", the criterion to use (e.g.: in, endswith, ==)
    method_content: Optional[str]
        If the type is "method", the code of the method to use.
        The method must be called `is_termination_message`,
        have one argument (`message`) which is a dict, and
        return a bool (whether the message is a termination message or not.)
    string : str
        The value of the termination message.

    Functions
    ---------
    validate_termination_message() -> Self
        Validate the termination message configuration.
    """

    type: Annotated[
        Literal["none", "keyword", "method"],
        Field(
            "none",
            title="Type",
            description=(
                "The type of the termination check to use: "
                "none, keyword, method"
            ),
        ),
    ]
    keywords: Annotated[
        List[str],
        Field(
            default_factory=list,
            title="Keywords",
            description=(
                "If the type is `keyword`,"
                " the keywords to search in the message."
            ),
        ),
    ]
    criterion: Annotated[
        Optional[Literal["found", "ending", "exact"]],
        Field(
            "exact",
            title="Criterion",
            description=(
                "If the type is `keyword`, "
                "the criterion to use (e.g.: in, endswith, ==)"
            ),
        ),
    ]
    method_content: Annotated[
        Optional[str],
        Field(
            None,
            title="Method content",
            description=(
                "If the type is `method`, the code of the method to use."
                "The method must be called `is_termination_message`,"
                "have one argument (`message`) which is a dict, and return a"
                "bool (whether the message is a termination message or not.)"
            ),
        ),
    ]

    _string: str = "None"

    def get_termination_function(
        self,
        name_prefix: Optional[str] = None,
        name_suffix: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Get the termination function.

        Parameters
        ----------
        name_prefix : str
            The function name prefix.
        name_suffix : str
            The function name suffix.

        Returns
        -------
        Tuple[str, str]
            The termination function and the function name.
        """
        function_name = "is_termination_message"
        if name_prefix:
            function_name = f"{name_prefix}_{function_name}"
        if name_suffix:
            function_name = f"{function_name}_{name_suffix}"
        if self.type in ("none", "keyword"):
            return self.string, function_name
        return (
            generate_function(
                function_name=function_name,
                function_args=IS_TERMINATION_MESSAGE_ARGS,
                function_types=IS_TERMINATION_MESSAGE_TYPES,
                function_body=self.string,
            ),
            function_name,
        )

    @property
    def string(self) -> str:
        """Get the value of the termination message.

        - If the type is "none", the value is "None".
        - If the type is "keyword", the value is a lambda function that checks
            if any of the keywords comply with the criterion.
        - If the type is "method", the value is the method content.

        Returns
        -------
        str
            The value of the termination message.
        """
        return self._string

    def _validate_method_content(self) -> None:
        """Validate the method content."""
        if not self.method_content:
            raise ValueError(
                "Method content is required for method termination."
            )
        valid, error_or_content = check_function(
            self.method_content,
            function_name=IS_TERMINATION_MESSAGE,
            function_args=IS_TERMINATION_MESSAGE_ARGS,
        )
        if not valid:
            raise ValueError(error_or_content)
        self._string = error_or_content

    def _validate_keyword(self) -> None:
        """Validate the keyword termination configuration."""
        if not self.keywords:
            raise ValueError("Keywords are required for keyword termination.")
        if self.criterion not in ["found", "ending", "exact"]:
            raise ValueError(f"Invalid criterion: {self.criterion}")
        # pylint: disable=inconsistent-quotes
        keywords_str = ", ".join([f'"{keyword}"' for keyword in self.keywords])
        if self.criterion == "found":
            self._string = (
                'lambda x: any(x.get("content", "") and '
                'keyword in x.get("content", "") '
                f"for keyword in [{keywords_str}])"
            )
        if self.criterion == "ending":
            self._string = (
                'lambda x: any(x.get("content", "") and '
                'x.get("content", "").endswith(keyword) '
                f"for keyword in [{keywords_str}])"
            )
        if self.criterion == "exact":
            self._string = (
                'lambda x: any(x.get("content", "") == keyword '
                f"for keyword in [{keywords_str}])"
            )

    @model_validator(mode="after")
    def validate_termination_message(self) -> Self:
        """Validate the termination message configuration.

        Raises
        ------
        ValueError
            If the configuration is invalid.
        Returns
        -------
        WaldiezAgentTerminationMessage
            The validated termination message configuration.
        """
        if self.type == "method":
            self._validate_method_content()
        if self.type == "keyword":
            self._validate_keyword()
        if self.type == "none":
            self._string = "None"
        return self



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get chroma db related imports and content."""

from pathlib import Path
from typing import Set, Tuple

from waldiez.models import WaldiezRagUser


def _get_chroma_client_string(agent: WaldiezRagUser) -> Tuple[str, str]:
    """Get the ChromaVectorDB client string.

    Parameters
    ----------
    agent : WaldiezRagUser
        The agent.
    agent_name : str
        The agent's name.

    Returns
    -------
    Tuple[str, str]
        The 'client' and what to import.
    """
    # TODO: also check `connection_url` (chroma in client-server mode)
    to_import = "import chromadb"
    client_str = "chromadb."
    if (
        agent.retrieve_config.db_config.use_local_storage
        and agent.retrieve_config.db_config.local_storage_path is not None
    ):
        # on windows, we might get:
        # SyntaxError: (unicode error) 'unicodeescape' codec can't decode bytes
        # in position 2-3: truncated \UXXXXXXXX escape
        local_path = Path(agent.retrieve_config.db_config.local_storage_path)
        client_str += (
            "PersistentClient(\n"
            f'    path=r"{local_path}",' + "\n"
            "    settings=Settings(anonymized_telemetry=False),\n"
            ")"
        )
    else:
        client_str += "Client(Settings(anonymized_telemetry=False))"
    return client_str, to_import


def _get_chroma_embedding_function_string(
    agent: WaldiezRagUser, agent_name: str
) -> Tuple[str, str, str]:
    """Get the ChromaVectorDB embedding function string.

    Parameters
    ----------
    agent : WaldiezRagUser
        The agent.
    agent_name : str
        The agent's name.

    Returns
    -------
    Tuple[str, str, str]
        The 'embedding_function', the import and the custom embedding function.
    """
    to_import = ""
    embedding_function_arg = ""
    embedding_function_content = ""
    vector_db_model = agent.retrieve_config.db_config.model
    if not agent.retrieve_config.use_custom_embedding:
        to_import = (
            "from chromadb.utils.embedding_functions "
            "import SentenceTransformerEmbeddingFunction"
        )
        embedding_function_arg = "SentenceTransformerEmbeddingFunction("
        embedding_function_arg += f'model_name="{vector_db_model}")'
    else:
        embedding_function_content, embedding_function_arg = (
            agent.retrieve_config.get_custom_embedding_function(
                name_suffix=agent_name
            )
        )
        embedding_function_content = "\n" + embedding_function_content
    return embedding_function_arg, to_import, embedding_function_content


def get_chroma_db_args(
    agent: WaldiezRagUser, agent_name: str
) -> Tuple[str, Set[str], str, str]:
    """Get the 'kwargs to use for ChromaVectorDB.

    Parameters
    ----------
    agent : WaldiezRagUser
        The agent.
    agent_name : str
        The agent's name.

    Returns
    -------
    Tuple[str, Set[str], str]

        - The 'kwargs' string.
        - What to import.
        - The custom embedding function.
        - Any additional content to be used before the `kwargs` string.
    """
    client_str, client_to_import = _get_chroma_client_string(agent)
    embedding_function_arg, to_import_embedding, embedding_function_body = (
        _get_chroma_embedding_function_string(agent, agent_name)
    )
    to_import = {client_to_import, "from chromadb.config import Settings"}
    if to_import_embedding:
        to_import.add(to_import_embedding)
    kwarg_string = (
        f"            client={agent_name}_client," + "\n"
        f"            embedding_function={embedding_function_arg}," + "\n"
    )
    # The RAG example:
    # https://ag2ai.github.io/ag2/docs/notebooks/agentchat_groupchat_RAG/
    # raises `InvalidCollectionException`: Collection groupchat does not exist.
    # https://github.com/chroma-core/chroma/issues/861
    # https://github.com/microsoft/autogen/issues/3551#issuecomment-2366930994
    # manually initializing the collection before running the flow,
    # might be a workaround.
    content_before = f"{agent_name}_client = {client_str}" + "\n"
    collection_name = agent.retrieve_config.collection_name
    get_or_create = agent.retrieve_config.get_or_create
    if collection_name:
        if get_or_create:
            content_before += (
                f"{agent_name}_client.get_or_create_collection("
                f'"{collection_name}")' + "\n"
            )
        else:
            content_before += (
                "try:\n"
                f'    {agent_name}_client.get_collection("{collection_name}")'
                + "\n"
                "except ValueError:\n"
                f"    {agent_name}_client.create_collection("
                f'"{collection_name}")' + "\n"
            )
    return kwarg_string, to_import, embedding_function_body, content_before



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Agent position generation."""

from dataclasses import dataclass
from enum import Enum
from typing import Optional

from waldiez.models import WaldiezAgent


class AgentPositions(Enum):
    """Agent positions.

    Attributes
    ----------
    BEFORE_ALL: int
        Before all agents.
    BEFORE: int
        Before the agent.
    AS_ARGUMENT: int
        As an argument of the agent's initialization.
    AFTER: int
        After the agent.
    AFTER_ALL: int
        After all agents.
    """

    BEFORE_ALL = 0
    BEFORE = 1
    AS_ARGUMENT = 2
    AFTER = 3
    AFTER_ALL = 4


POSITIONS_WITHOUT_AGENT = [
    AgentPositions.BEFORE_ALL,
    AgentPositions.AFTER_ALL,
]


@dataclass(order=True, frozen=True, slots=True)
class AgentPosition:
    """Agent position.

    Attributes
    ----------
    agent: Optional[WaldiezAgent]
        The agent.
    position: AgentPositions
        The position.
    order: int
        The order of the agent position.

    Raises
    ------
    ValueError
        If the position is not "BEFORE_ALL" or "AFTER_ALL"
        and the agent is not provided.
    """

    agent: Optional[WaldiezAgent]
    position: AgentPositions
    order: int = 0

    def __post_init__(self) -> None:
        """Post initialization.

        Raises
        ------
        ValueError
            If the agent is not provided for the given position.
        """
        if self.agent is None and self.position not in POSITIONS_WITHOUT_AGENT:
            raise ValueError("Agent must be provided for the given position.")



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Reasoning agent's reason configuration model."""

from pydantic import Field
from typing_extensions import Annotated, Literal

from ...common import WaldiezBase

ReasoningConfigMethod = Literal["beam_search", "mcts", "lats", "dfs"]


class WaldiezReasoningAgentReasonConfig(WaldiezBase):
    """Reasoning agent's reason configuration model.

    Configuration for the reasoning method.

    Attributes
    ----------
    method : Literal["beam_search", "mcts", "lats", "dfs"]
        The search strategy to use, default is "beam_search".
    max_depth : int
        Maximum depth of reasoning tree, default is 3.
    forest_size : int
        Number of independent trees to maintain, default is 1.
    rating_scale : int
        Scale for grading responses, e.g. 1-10, default is 10.
    beam_size : int
        Number of parallel paths to maintain, default is 3 (for beam_search).
    answer_approach : Literal["pool", "best"]
        How to select final answer, default is "pool" (only for beam_search).
    nsim : int
        Number of simulations to run, default is 3 (only for mcts and lats).
    exploration_constant : float
        UCT exploration parameter, default is 1.41 (only for mcts and lats).
    """

    method: Annotated[
        ReasoningConfigMethod,
        Field(
            "beam_search",
            title="Method",
            description="The search strategy to use.",
        ),
    ]
    max_depth: Annotated[
        int,
        Field(
            3,
            title="Maximum depth",
            description="Maximum depth of reasoning tree.",
        ),
    ]
    forest_size: Annotated[
        int,
        Field(
            1,
            title="Forest size",
            description="Number of independent trees to maintain.",
        ),
    ]
    rating_scale: Annotated[
        int,
        Field(
            10,
            title="Rating scale",
            description="Scale for grading responses, e.g. 1-10.",
        ),
    ]
    beam_size: Annotated[
        int,
        Field(
            3,
            title="Beam size",
            description="Number of parallel paths to maintain.",
        ),
    ]
    answer_approach: Annotated[
        Literal["pool", "best"],
        Field(
            "pool",
            title="Answer approach",
            description="How to select final answer.",
        ),
    ]
    nsim: Annotated[
        int,
        Field(
            3,
            title="Number of simulations",
            description="Number of simulations to run.",
        ),
    ]
    exploration_constant: Annotated[
        float,
        Field(
            1.41,
            title="Exploration constant",
            description="UCT exploration parameter.",
        ),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Try to suppress the annoying flaml.automl not being available warning."""

import logging

__WALDIEZ_CHECKED_FLAML_WARNINGS = False


def check_flaml_warnings() -> None:  # pragma: no cover
    """Check for flaml warnings once."""
    # pylint: disable=global-statement
    global __WALDIEZ_CHECKED_FLAML_WARNINGS
    if __WALDIEZ_CHECKED_FLAML_WARNINGS is False:
        flam_logger = logging.getLogger("flaml")
        flam_logger.setLevel(logging.ERROR)
        __WALDIEZ_CHECKED_FLAML_WARNINGS = True



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: skip-file
# isort: skip_file
"""Extra typer commands for CLI."""

import typer

from .jupyter import add_jupyter_cli
from .studio import add_studio_cli


def add_cli_extras(app: typer.Typer) -> None:
    """Add extra CLI commands to the app.

    Parameters
    ----------
    app : typer.Typer
        The Typer app to add the extra commands to.

    Returns
    -------
    typer.Typer
        The app with the extra commands added
    """
    add_jupyter_cli(app)
    add_studio_cli(app)


__all__ = ["add_cli_extras"]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utilities for naming.

Functions
---------
get_valid_python_variable_name
    Make sure a string is a valid Python variable name.
get_valid_instance_name
    Get a valid instance name.
get_escaped_string
    Get a string with escaped quotes and newlines.
"""

import re
from typing import Dict, Tuple


def get_valid_python_variable_name(
    possible: str,
    prefix: str = "w",
    max_length: int = 64,
) -> str:
    """Get a valid Python variable name from a possible name.

    Parameters
    ----------
    possible : str
        The possible name.

    prefix : str, optional
        The prefix to use if the name starts with a digit or special character

    max_length : int, optional
        The maximum length of the variable name.

    Returns
    -------
    str
        The valid Python variable name.
    """

    def replacement(match: re.Match[str]) -> str:
        """Get the replacement for the match.

        Parameters
        ----------
        match : re.Match[str]
            The match.

        Returns
        -------
        str
            The replacement
        """
        if match.group(0) in ["->", "=>"]:
            return "to"
        if match.group(0) in ["<-", "<="]:
            return "from"
        if re.match(r"\W|^(?=\d)", match.group(0)):
            return "_"
        return match.group(0)

    possible = re.sub(r"->|=>|<-|<=|\W|^(?=\d)", replacement, possible)[
        :max_length
    ].lower()

    if not possible:
        return prefix + "_"
    if possible.startswith("_"):
        return f"{prefix}{possible}"
    if possible[0].isdigit():
        return f"{prefix}_{possible}"
    return possible


def get_valid_instance_name(
    instance: Tuple[str, str],
    current_names: Dict[str, str],
    prefix: str = "w",
    max_length: int = 64,
) -> Dict[str, str]:
    """Get a valid instance name.

    If the instance id is already in the current names nothing is done.
    If the name already exists in the current names,
        the name is updated (with an index suffix).

    Parameters
    ----------
    instance : Tuple[str, str]
        The instance id and possible name.
    current_names : Dict[str, str]
        The current names.
    prefix : str, optional
        The prefix to use if the name starts with a digit,
        if the name is already in the current names,
        or if the name is already in the current names with an index suffix.
    max_length : int, optional
        The maximum length of the variable name.

    Returns
    -------
    Dict[str, str]
        The updated names.
    """
    instance_id, possible_name = instance[0], instance[1][:max_length]
    if instance_id in current_names:
        # already in the current names (it's id)
        return current_names
    new_names = current_names.copy()
    name = get_valid_python_variable_name(possible_name, prefix=prefix)
    if name in current_names.values():
        name = f"{prefix}_{name}"
    if name in current_names.values():
        index = 1
        while f"{name}_{index}" in current_names.values():
            index += 1
        name = f"{name}_{index}"
    new_names[instance_id] = name
    return new_names


def get_escaped_string(string: str) -> str:
    """Get a string with escaped quotes and newlines.

    Parameters
    ----------
    string : str
        The original string.

    Returns
    -------
    str
        The escaped string.
    """
    return string.replace('"', '\\"').replace("\n", "\\n")



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Reasoning agent data model."""

from typing import Any, Dict

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..assistant import WaldiezAssistantData
from .reasoning_agent_reason_config import WaldiezReasoningAgentReasonConfig


class WaldiezReasoningAgentData(WaldiezAssistantData):
    """Reasoning agent data model."""

    max_depth: Annotated[
        int,
        Field(
            4,
            title="Maximum depth",
            description="Maximum depth of the reasoning tree",
            alias="maxDepth",
            deprecated=True,
        ),
    ] = 4
    beam_size: Annotated[
        int,
        Field(
            3,
            title="Beam size",
            description="Number of parallel reasoning paths to maintain",
            alias="beamSize",
            deprecated=True,
        ),
    ] = 3
    answer_approach: Annotated[
        Literal["pool", "best"],
        Field(
            "pool",
            title="Answer approach",
            description="How to generate final answer",
            alias="answerApproach",
            deprecated=True,
        ),
    ] = "pool"
    verbose: Annotated[
        bool,
        Field(
            True,
            title="Verbose",
            description="Whether to show intermediate steps",
        ),
    ]
    reason_config: Annotated[
        WaldiezReasoningAgentReasonConfig,
        Field(
            title="Reason config",
            description="The reasoning agent's reason configuration",
            default_factory=WaldiezReasoningAgentReasonConfig,
            alias="reasonConfig",
        ),
    ]

    def get_reasoning_config(self) -> Dict[str, Any]:
        """Get the reasoning configuration based on the reason_config method.

        Returns
        -------
        Dict[str, Any]
            The reasoning configuration.

        """
        reason_dict: Dict[str, Any] = {
            "method": self.reason_config.method,
            "max_depth": self.reason_config.max_depth,
            "forest_size": self.reason_config.forest_size,
            "rating_scale": self.reason_config.rating_scale,
        }
        if self.reason_config.method == "beam_search":
            reason_dict["beam_size"] = self.reason_config.beam_size
            reason_dict["answer_approach"] = self.reason_config.answer_approach
        if self.reason_config.method in ("mcts", "lats"):
            reason_dict["nsim"] = self.reason_config.nsim
            reason_dict["exploration_constant"] = (
                self.reason_config.exploration_constant
            )
        return reason_dict


# reason_config (dict): Configuration for the reasoning method.
# Supported parameters:
#     method (str): The search strategy to use. Options:
#         - "beam_search" (default): Uses beam search with parallel paths
#         - "mcts": Uses Monte Carlo Tree Search for exploration
#         - "lats": Uses Language Agent Tree Search with per-step rewards
#         - "dfs": Uses depth-first search
#                (equivalent to beam_search with beam_size=1)
#     Common parameters:
#         max_depth (int): Maximum depth of reasoning tree (default: 3)
#         forest_size (int):
#               Number of independent trees to maintain (default: 1)
#         rating_scale (int):
#               Scale for grading responses, e.g. 1-10 (default: 10)
#     Beam Search specific:
#         beam_size (int): Number of parallel paths to maintain (default: 3)
#         answer_approach (str):
#               How to select final answer, "pool" or "best" (default: "pool")
#     MCTS/LATS specific:
#         nsim (int): Number of simulations to run (default: 3)
#         exploration_constant (float):
#               UCT exploration parameter (default: 1.41)
#     Example configs:
#         `{"method": "beam_search", "beam_size": 5, "max_depth": 4}`
#         `{"method": "mcts", "nsim": 10, "exploration_constant": 2.0}`
#         `{"method": "lats", "nsim": 5, "forest_size": 3}`



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utility functions for exporting waldiez to different formats."""

from .agent_utils import (
    add_after_agent_content,
    add_after_all_agents_content,
    add_before_agent_content,
    add_before_all_agents_content,
    gather_agent_outputs,
)
from .chat_utils import add_after_chat_content, add_before_chat_content
from .def_main import get_def_main
from .flow_content import (
    get_after_run_content,
    get_ipynb_content_start,
    get_py_content_start,
)
from .flow_names import ensure_unique_names
from .importing_utils import (
    gather_imports,
    get_standard_imports,
    get_the_imports_string,
)
from .logging_utils import (
    get_sqlite_out,
    get_sqlite_out_call,
    get_start_logging,
    get_stop_logging,
)

__all__ = [
    "add_after_agent_content",
    "add_after_all_agents_content",
    "add_before_agent_content",
    "add_before_all_agents_content",
    "add_after_chat_content",
    "add_before_chat_content",
    "ensure_unique_names",
    "gather_agent_outputs",
    "gather_imports",
    "get_after_run_content",
    "get_def_main",
    "get_py_content_start",
    "get_ipynb_content_start",
    "get_start_logging",
    "get_stop_logging",
    "get_sqlite_out",
    "get_sqlite_out_call",
    "get_standard_imports",
    "get_the_imports_string",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Group chat manager data."""

from typing import Optional

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..agent import WaldiezAgentData
from .speakers import WaldiezGroupManagerSpeakers


class WaldiezGroupManagerData(WaldiezAgentData):
    """Group chat manager data class.

    The data for an agent with `human_input_mode` set to "NEVER" as default.
    and the chat group's related extra properties.
    See the parent's docs (`WaldiezAgentData`) for the rest of the properties.

    Attributes
    ----------
    human_input_mode : Literal["ALWAYS", "NEVER", "TERMINATE"]
        The human input mode, Defaults to `NEVER`
    max_round : Optional[int]
        The maximum number of rounds to have in the group.
    admin_name : Optional[str]
        The name of the group's admin.
        Make sure you use a name of an agent in the group.
    speakers : WaldiezGroupManagerSpeakers
        The rules for the speaker selection and repetition
    enable_clear_history : Optional[bool]
        Enable clearing the history in the chat group.
    send_introductions : bool
        Send the group members' introductions.
    """

    human_input_mode: Annotated[
        Literal["ALWAYS", "NEVER", "TERMINATE"],
        Field(
            "NEVER",
            title="Human input mode",
            description="The human input mode, Defaults to `NEVER`",
            alias="humanInputMode",
        ),
    ]
    max_round: Annotated[
        Optional[int],
        Field(
            None,
            title="Max round",
            description="The maximum number of rounds to have in the group.",
            alias="maxRound",
        ),
    ]
    admin_name: Annotated[
        Optional[str],
        Field(
            None,
            title="Group Admin name",
            description=(
                "The name of the group's admin. "
                "Make sure you use a name of an agent in the group."
            ),
            alias="adminName",
        ),
    ]
    speakers: Annotated[
        WaldiezGroupManagerSpeakers,
        Field(
            title="Speakers",
            description="The rules for the speaker selection and repetition",
            default_factory=WaldiezGroupManagerSpeakers,
        ),
    ]
    enable_clear_history: Annotated[
        Optional[bool],
        Field(
            None,
            title="Enable clear history",
            description="Enable clearing hte history in the chat group.",
            alias="enableClearHistory",
        ),
    ]
    send_introductions: Annotated[
        bool,
        Field(
            False,
            title="Send Introductions",
            description="Send the group members' introductions.",
            alias="sendIntroductions",
        ),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""It extends a user agent and has RAG related parameters."""

from .rag_user import WaldiezRagUser
from .rag_user_data import WaldiezRagUserData
from .retrieve_config import (
    CUSTOM_EMBEDDING_FUNCTION,
    CUSTOM_EMBEDDING_FUNCTION_ARGS,
    CUSTOM_EMBEDDING_FUNCTION_TYPES,
    CUSTOM_TEXT_SPLIT_FUNCTION,
    CUSTOM_TEXT_SPLIT_FUNCTION_ARGS,
    CUSTOM_TEXT_SPLIT_FUNCTION_TYPES,
    CUSTOM_TOKEN_COUNT_FUNCTION,
    CUSTOM_TOKEN_COUNT_FUNCTION_ARGS,
    CUSTOM_TOKEN_COUNT_FUNCTION_TYPES,
    WaldiezRagUserChunkMode,
    WaldiezRagUserModels,
    WaldiezRagUserRetrieveConfig,
    WaldiezRagUserTask,
    WaldiezRagUserVectorDb,
)
from .vector_db_config import WaldiezRagUserVectorDbConfig

__all__ = [
    "CUSTOM_EMBEDDING_FUNCTION",
    "CUSTOM_EMBEDDING_FUNCTION_ARGS",
    "CUSTOM_EMBEDDING_FUNCTION_TYPES",
    "CUSTOM_TEXT_SPLIT_FUNCTION",
    "CUSTOM_TEXT_SPLIT_FUNCTION_ARGS",
    "CUSTOM_TEXT_SPLIT_FUNCTION_TYPES",
    "CUSTOM_TOKEN_COUNT_FUNCTION",
    "CUSTOM_TOKEN_COUNT_FUNCTION_ARGS",
    "CUSTOM_TOKEN_COUNT_FUNCTION_TYPES",
    "WaldiezRagUser",
    "WaldiezRagUserData",
    "WaldiezRagUserModels",
    "WaldiezRagUserVectorDb",
    "WaldiezRagUserChunkMode",
    "WaldiezRagUserRetrieveConfig",
    "WaldiezRagUserTask",
    "WaldiezRagUserVectorDbConfig",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Common utils for all models."""

from .base import WaldiezBase
from .date_utils import now
from .dict_utils import update_dict
from .method_utils import (
    check_function,
    generate_function,
    get_function,
    parse_code_string,
)

__all__ = [
    "WaldiezBase",
    "now",
    "check_function",
    "get_function",
    "generate_function",
    "parse_code_string",
    "update_dict",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utils for exporting chats."""

from .nested import export_nested_chat_registration
from .sequential import export_sequential_chat
from .single_chat import export_single_chat
from .swarm import export_swarm_chat

__all__ = [
    "export_nested_chat_registration",
    "export_sequential_chat",
    "export_single_chat",
    "export_swarm_chat",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Tools for exporting agents, models, skills and chats to strings."""

from .agent import AgentExporter
from .flow import FlowExporter
from .models import ModelsExporter
from .skills import SkillsExporter

__all__ = [
    "AgentExporter",
    "FlowExporter",
    "ModelsExporter",
    "SkillsExporter",
]



"""Date utilities."""

from datetime import datetime, timezone


def now() -> str:
    """Get the current date and time in UTC.

    Returns
    -------
    str
        The current date and time in UTC.
    """
    return (
        datetime.now(tz=timezone.utc)
        .isoformat(timespec="milliseconds")
        .replace("+00:00", "Z")
    )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa E501

"""The vector db config for the RAG user agent."""

from pathlib import Path
from typing import Any, Dict, Optional

from pydantic import ConfigDict, Field, model_validator
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated, Self

from ...common import WaldiezBase


# pylint: disable=line-too-long
class WaldiezRagUserVectorDbConfig(WaldiezBase):
    """The config for the vector db.

    Attributes
    ----------
    model : str
        The model to use for the vector db embeddings.
    use_memory : bool
        Whether to use memory for the vector db (if `qdrant` is used).
    use_local_storage : bool
        Whether to use local storage for the db (if `qdrant` or `chroma` is used).
    local_storage_path : Optional[str]
        The path to the local storage for the vector db (if `qdrant` or `chroma` is used).
    connection_url : Optional[str]
        The connection url for the vector db.
    wait_until_index_ready : Optional[float]
        Blocking call to wait until the database indexes are ready (if `mongodb` is used).
        None, the default, means no wait.
    wait_until_document_ready : Optional[float]
        Blocking call to wait until the database documents are ready (if `mongodb` is used).
        None, the default, means no wait.
    metadata : Optional[Dict[str, Any]]
        The metadata to use for the vector db.
        Example: {"hnsw:space": "ip", "hnsw:construction_ef": 30, "hnsw:M": 32}

    Functions
    ---------
    validate_vector_db_config()
        Validate the vector db config.
    """

    model_config = ConfigDict(
        extra="ignore",
        alias_generator=to_camel,
        populate_by_name=True,
        frozen=False,
        protected_namespaces=(),  # we use "model" as a field
    )

    model: Annotated[
        Optional[str],
        Field(
            None,
            title="Model",
            description="The model to use for the vector db embeddings.",
        ),
    ]
    use_memory: Annotated[
        bool,
        Field(
            True,
            title="Use Memory",
            description=(
                "Whether to use memory for the vector db (if qdrant is used)."
            ),
        ),
    ]
    use_local_storage: Annotated[
        bool,
        Field(
            False,
            title="Use Local Storage",
            description=(
                "Whether to use local storage for the vector db "
                "(if qdrant or chroma is used)."
            ),
        ),
    ]
    local_storage_path: Annotated[
        Optional[str],
        Field(
            None,
            title="Local Storage Path",
            description=(
                "The path to the local storage for the vector db "
                "(if qdrant is used)."
            ),
        ),
    ]
    connection_url: Annotated[
        Optional[str],
        Field(
            None,
            title="Connection URL",
            description="The connection url for the vector db.",
        ),
    ]
    wait_until_index_ready: Annotated[
        Optional[float],
        Field(
            None,
            title="Wait Until Index Ready",
            description=(
                "The time to wait/block until the database indexes are ready. "
                "None, the default, means no wait."
            ),
        ),
    ]
    wait_until_document_ready: Annotated[
        Optional[float],
        Field(
            None,
            title="Wait Until Document Ready",
            description=(
                "The time to wait/block until the database documents "
                "are ready. None, the default, means no wait."
            ),
        ),
    ]
    metadata: Annotated[
        Optional[Dict[str, Any]],
        Field(
            None,
            title="Metadata",
            description=(
                "The metadata to use for the vector db. Example: "
                '"{"hnsw:space": "ip", '
                '"hnsw:construction_ef": 30, '
                '"hnsw:M": 32}'
            ),
        ),
    ]

    @model_validator(mode="after")
    def validate_vector_db_config(self) -> Self:
        """Validate the vector db config.

        if local storage is used, make sure the path is provided,
        and make it absolute if not already.

        Returns
        -------
        WaldiezRagUserVectorDbConfig
            The vector db config.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if self.use_local_storage:
            if self.local_storage_path is None:
                raise ValueError(
                    "The local storage path must be provided if local storage is used."
                )
            as_path = Path(self.local_storage_path)
            if not as_path.is_absolute():
                self.local_storage_path = str(as_path.resolve())
        return self



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez flow data."""

from typing import Any, Dict, List

from pydantic import Field, model_validator
from typing_extensions import Annotated, Self

from ..agents import WaldiezAgents
from ..chat import WaldiezChat
from ..common import WaldiezBase
from ..model import WaldiezModel
from ..skill import WaldiezSkill


class WaldiezFlowData(WaldiezBase):
    """Flow data class.

    Attributes
    ----------
    nodes : List[Dict[str, Any]]
        The nodes of the flow. We ignore this (UI-related)
    edges : List[Dict[str, Any]]
        The edges of the flow. We ignore this (UI-related)
    viewport : Dict[str, Any]
        The viewport of the flow. We ignore this (UI-related)
    agents : WaldiezAgents
        The agents of the flow:
        users: List[WaldiezUserProxy]
        assistants: List[WaldiezAssistant]
        managers: List[WaldiezGroupManager]
        rag_users : List[WaldiezRagUser]
        See `WaldiezAgents` for more info.
    models : List[WaldiezModel]
        The models of the flow. See `WaldiezModel`.
    skills : List[WaldiezSkill]
        The skills of the flow. See `WaldiezSkill`.
    chats : List[WaldiezChat]
        The chats of the flow. See `WaldiezChat`.
    is_async : bool
        Whether the flow is asynchronous or not.
    """

    # the ones below (nodes,edges, viewport) we ignore
    # (they for graph connections, positions, etc.)
    nodes: Annotated[
        List[Dict[str, Any]],
        Field(
            default_factory=list,
            title="Nodes",
            description="The nodes of the flow",
        ),
    ]
    edges: Annotated[
        List[Dict[str, Any]],
        Field(
            default_factory=list,
            title="Edges",
            description="The edges of the flow",
        ),
    ]
    viewport: Annotated[
        Dict[str, Any],
        Field(
            default_factory=dict,
            title="Viewport",
            description="The viewport of the flow",
        ),
    ]
    # these are the ones we use.
    agents: Annotated[
        WaldiezAgents,
        Field(
            description="The agents of the flow",
            title="Agents",
            default_factory=WaldiezAgents,
        ),
    ]
    models: Annotated[
        List[WaldiezModel],
        Field(
            description="The models of the flow",
            title="Models",
            default_factory=list,
        ),
    ]
    skills: Annotated[
        List[WaldiezSkill],
        Field(
            description="The skills of the flow",
            title="Skills",
            default_factory=list,
        ),
    ]
    chats: Annotated[
        List[WaldiezChat],
        Field(
            description="The chats of the flow",
            title="Chats",
            default_factory=list,
        ),
    ]
    is_async: Annotated[
        bool,
        Field(
            False,
            description="Whether the flow is asynchronous or not",
            title="Is Async",
        ),
    ]

    @model_validator(mode="after")
    def validate_flow_chats(self) -> Self:
        """Validate the flow chats.

        Returns
        -------
        WaldiezFlowData
            The flow data.

        Raises
        ------
        ValueError
            If there is a chat with a prerequisite that does not exist.
        """
        self.chats = sorted(self.chats, key=lambda x: x.order)
        # in async, ag2 uses the "chat_id" field (and it must be an int):
        # ```
        #    prerequisites = []
        #    for chat_info in chat_queue:
        #        if "chat_id" not in chat_info:
        #            raise ValueError(
        #               "Each chat must have a unique id for "
        #               "async multi-chat execution."
        #            )
        #     chat_id = chat_info["chat_id"]
        #     pre_chats = chat_info.get("prerequisites", [])
        #     for pre_chat_id in pre_chats:
        #         if not isinstance(pre_chat_id, int):
        #             raise ValueError("Prerequisite chat id is not int.")
        #         prerequisites.append((chat_id, pre_chat_id))
        #    return prerequisites
        # ```
        id_to_chat_id: Dict[str, int] = {}
        for index, chat in enumerate(self.chats):
            id_to_chat_id[chat.id] = index
            chat.set_chat_id(index)
        # also update the chat prerequisites
        #  we have ids(str), not chat_ids(int)
        for chat in self.chats:
            chat_prerequisites = []
            for chat_id in chat.data.prerequisites:
                if chat_id not in id_to_chat_id:  # pragma: no cover
                    raise ValueError(
                        f"Chat with id {chat_id} not found in the flow."
                    )
                chat_prerequisites.append(id_to_chat_id[chat_id])
            chat.set_prerequisites(chat_prerequisites)
        return self
